# Ensure coverage starts for all Python processes so that test coverage is calculated
# properly when using subprocesses (see https://coverage.readthedocs.io/en/latest/subprocess.html)
import coverage
print("Starting coverage")
coverage.process_startup()



import requests
from ..logger import logger


class SocialSlack(object):

    """Social Messaging sink to output to Slack."""

    def __init__(self, **kwargs):
        self.web_hook = kwargs.get('webhook_url', '')
        if self.web_hook == '':
            raise ValueError('webhook_url parameter is not defined.')
        else:
            self.output_timestamp = kwargs.get('output_timestamp', False)

    def send_message(self, msg, timestamp):
        try:
            if self.output_timestamp:
                post_msg = '{} - {}'.format(msg, timestamp)
            else:
                post_msg = msg

            # We ignore the response body and headers of a successful post.
            requests.post(self.web_hook, json={'text': post_msg})
        except Exception as e:  # pragma: no cover
            logger.warning('Error posting to slack: {}'.format(e))



# This module implements a handler for serial_for_url("no_op://").

from panoptes.utils.tests.serial_handlers import NoOpSerial

# Export it as Serial so that it will be picked up by PySerial's serial_for_url.
Serial = NoOpSerial



import os
import re
import subprocess
import shutil
from contextlib import suppress
from warnings import warn

import numpy as np
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.figure import Figure

from astropy.wcs import WCS
from astropy.nddata import Cutout2D
from astropy.visualization import (PercentileInterval, LogStretch, ImageNormalize)

from dateutil import parser as date_parser

from .. import error
from ..logger import logger
from ..time import current_time
from ..images import fits as fits_utils
from ..images.plot import add_colorbar
from ..images.plot import get_palette


def crop_data(data, box_width=200, center=None, data_only=True, wcs=None, **kwargs):
    """Return a cropped portion of the image

    Shape is a box centered around the middle of the data

    Args:
        data (`numpy.array`): Array of data.
        box_width (int, optional): Size of box width in pixels, defaults to 200px.
        center (tuple(int, int), optional): Crop around set of coords, default to image center.
        data_only (bool, optional): If True (default), return only data. If False
            return the `Cutout2D` object.
        wcs (None|`astropy.wcs.WCS`, optional): A valid World Coordinate System (WCS) that will
            be cropped along with the data if provided.

    Returns:
        np.array: A clipped (thumbnailed) version of the data if `data_only=True`, otherwise
            a `astropy.nddata.Cutout2D` object.

    """
    assert data.shape[0] >= box_width, "Can't clip data, it's smaller than {} ({})".format(
        box_width, data.shape)
    # Get the center
    if center is None:
        x_len, y_len = data.shape
        x_center = int(x_len / 2)
        y_center = int(y_len / 2)
    else:
        y_center = int(center[0])
        x_center = int(center[1])

    logger.debug("Using center: {} {}".format(x_center, y_center))
    logger.debug("Box width: {}".format(box_width))

    cutout = Cutout2D(data, (y_center, x_center), box_width, wcs=wcs)

    if data_only:
        return cutout.data

    return cutout


def make_pretty_image(fname,
                      title=None,
                      timeout=15,
                      img_type=None,
                      link_path=None,
                      **kwargs):
    """Make a pretty image.

    This will create a jpg file from either a CR2 (Canon) or FITS file.

    Notes:
        See `/scripts/cr2_to_jpg.sh` for CR2 process.

    Arguments:
        fname (str): The path to the raw image.
        title (None|str, optional): Title to be placed on image, default None.
        timeout (int, optional): Timeout for conversion, default 15 seconds.
        img_type (None|str, optional): Image type of fname, one of '.cr2' or '.fits'.
            The default is `None`, in which case the file extension of fname is used.
        link_path (None|str, optional): Path to location that image should be symlinked.
            The directory must exist.
        **kwargs {dict} -- Additional arguments to be passed to external script.

    Returns:
        str -- Filename of image that was created.

    Deleted Parameters:
        link_latest (bool, optional): If the pretty picture should be linked to
            `/images/latest.jpg`, default False.
    """
    if img_type is None:
        img_type = os.path.splitext(fname)[-1]

    if not os.path.exists(fname):
        warn("File doesn't exist, can't make pretty: {}".format(fname))
        return None
    elif img_type == '.cr2':
        pretty_path = _make_pretty_from_cr2(fname, title=title, timeout=timeout, **kwargs)
    elif img_type in ['.fits', '.fz']:
        pretty_path = _make_pretty_from_fits(fname, title=title, **kwargs)
    else:
        warn("File must be a Canon CR2 or FITS file.")
        return None

    if link_path is None or not os.path.exists(os.path.dirname(link_path)):
        return pretty_path

    # Remove existing symlink
    with suppress(FileNotFoundError):
        os.remove(link_path)

    try:
        os.symlink(pretty_path, link_path)
    except Exception as e:
        warn("Can't link latest image: {}".format(e))

    return link_path


def _make_pretty_from_fits(fname=None,
                           title=None,
                           figsize=(10, 10 / 1.325),
                           dpi=150,
                           alpha=0.2,
                           number_ticks=7,
                           clip_percent=99.9,
                           **kwargs):

    data = mask_saturated(fits_utils.getdata(fname))
    header = fits_utils.getheader(fname)
    wcs = WCS(header)

    if not title:
        field = header.get('FIELD', 'Unknown field')
        exptime = header.get('EXPTIME', 'Unknown exptime')
        filter_type = header.get('FILTER', 'Unknown filter')

        try:
            date_time = header['DATE-OBS']
        except KeyError:
            # If we don't have DATE-OBS, check filename for date
            try:
                basename = os.path.splitext(os.path.basename(fname))[0]
                date_time = date_parser.parse(basename).isoformat()
            except Exception:
                # Otherwise use now
                date_time = current_time(pretty=True)

        date_time = date_time.replace('T', ' ', 1)

        title = '{} ({}s {}) {}'.format(field, exptime, filter_type, date_time)

    norm = ImageNormalize(interval=PercentileInterval(clip_percent), stretch=LogStretch())

    fig = Figure()
    FigureCanvas(fig)
    fig.set_size_inches(*figsize)
    fig.dpi = dpi

    if wcs.is_celestial:
        ax = fig.add_subplot(1, 1, 1, projection=wcs)
        ax.coords.grid(True, color='white', ls='-', alpha=alpha)

        ra_axis = ax.coords['ra']
        ra_axis.set_axislabel('Right Ascension')
        ra_axis.set_major_formatter('hh:mm')
        ra_axis.set_ticks(
            number=number_ticks,
            color='white',
            exclude_overlapping=True
        )

        dec_axis = ax.coords['dec']
        dec_axis.set_axislabel('Declination')
        dec_axis.set_major_formatter('dd:mm')
        dec_axis.set_ticks(
            number=number_ticks,
            color='white',
            exclude_overlapping=True
        )
    else:
        ax = fig.add_subplot(111)
        ax.grid(True, color='white', ls='-', alpha=alpha)

        ax.set_xlabel('X / pixels')
        ax.set_ylabel('Y / pixels')

    im = ax.imshow(data, norm=norm, cmap=get_palette(), origin='lower')
    add_colorbar(im)
    fig.suptitle(title)

    new_filename = re.sub('.fits(.fz)?', '.jpg', fname)
    fig.savefig(new_filename, bbox_inches='tight')

    # explicitly close and delete figure
    fig.clf()
    del fig

    return new_filename


def _make_pretty_from_cr2(fname, title=None, timeout=15, **kwargs):
    script_name = shutil.which('cr2-to-jpg')
    cmd = [script_name, fname]

    if title:
        cmd.append(title)

    logger.debug(cmd)

    try:
        output = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        logger.debug(output)
    except Exception as e:
        raise error.InvalidCommand(f"Error executing {script_name}: {e.output!r}\nCommand: {cmd}")

    return fname.replace('cr2', 'jpg')


def mask_saturated(data, saturation_level=None, threshold=0.9, dtype=np.float64):
    """Convert data to masked array of requested dtype with saturated values masked.

    Args:
        data (array_like): The numpy data array.
        saturation_level (float|None, optional): The saturation level. If None,
            the level will be set to the threshold times the max value for the dtype.
        threshold (float, optional): The percentage of the max value to use.
        dtype (`numpy.dtype`, optional): The requested dtype for the new array.

    Returns:
        `numpy.ma.array`: The masked numpy array.
    """
    if not saturation_level:
        try:
            # If data is an integer type use iinfo to compute machine limits
            dtype_info = np.iinfo(data.dtype)
        except ValueError:
            # Not an integer type. Assume for now we have 16 bit data
            saturation_level = threshold * (2**16 - 1)
        else:
            # Data is an integer type, set saturation level at specified fraction of
            # max value for the type
            saturation_level = threshold * dtype_info.max

    # Convert data to masked array of requested dtype, mask values above saturation level
    return np.ma.array(data, mask=(data > saturation_level), dtype=dtype)


def make_timelapse(
        directory,
        fn_out=None,
        glob_pattern='20[1-9][0-9]*T[0-9]*.jpg',
        overwrite=False,
        timeout=60,
        **kwargs):
    """Create a timelapse.

    A timelapse is created from all the images in a given `directory`

    Args:
        directory (str): Directory containing image files
        fn_out (str, optional): Full path to output file name, if not provided,
            defaults to `directory` basename.
        glob_pattern (str, optional): A glob file pattern of images to include,
            default '20[1-9][0-9]*T[0-9]*.jpg', which corresponds to the observation
            images but excludes any pointing images. The pattern should be relative
            to the local directory.
        overwrite (bool, optional): Overwrite timelapse if exists, default False.
        timeout (int): Timeout for making movie, default 60 seconds.
        **kwargs (dict):

    Returns:
        str: Name of output file

    Raises:
        error.InvalidSystemCommand: Raised if ffmpeg command is not found.
        FileExistsError: Raised if fn_out already exists and overwrite=False.
    """
    if fn_out is None:
        head, tail = os.path.split(directory)
        if tail == '':
            head, tail = os.path.split(head)

        field_name = head.split('/')[-2]
        cam_name = head.split('/')[-1]
        fname = '{}_{}_{}.mp4'.format(field_name, cam_name, tail)
        fn_out = os.path.normpath(os.path.join(directory, fname))

    if os.path.exists(fn_out) and not overwrite:
        raise FileExistsError("Timelapse exists. Set overwrite=True if needed")

    ffmpeg = shutil.which('ffmpeg')
    if ffmpeg is None:
        raise error.InvalidSystemCommand("ffmpeg not found, can't make timelapse")

    inputs_glob = os.path.join(directory, glob_pattern)

    try:
        ffmpeg_cmd = [
            ffmpeg,
            '-r', '3',
            '-pattern_type', 'glob',
            '-i', inputs_glob,
            '-s', 'hd1080',
            '-vcodec', 'libx264',
        ]

        if overwrite:
            ffmpeg_cmd.append('-y')

        ffmpeg_cmd.append(fn_out)

        logger.debug(ffmpeg_cmd)

        proc = subprocess.Popen(ffmpeg_cmd, universal_newlines=True,
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        try:
            # Don't wait forever
            outs, errs = proc.communicate(timeout=timeout)
        except subprocess.TimeoutExpired:
            proc.kill()
            outs, errs = proc.communicate()
        finally:
            logger.debug(f"Output: {outs}")
            logger.debug(f"Errors: {errs}")

            # Double-check for file existence
            if not os.path.exists(fn_out):
                fn_out = None
    except Exception as e:
        raise error.PanError(f"Problem creating timelapse in {fn_out}: {e!r}")

    return fn_out



import pytest

from panoptes.utils.database import PanDB
from panoptes.utils import error


def test_bad_db():
    with pytest.raises(Exception):
        PanDB('foobar')


def test_insert_and_no_permanent(db):
    rec = {'test': 'insert'}
    id0 = db.insert_current('config', rec, store_permanently=False)

    record = db.get_current('config')
    assert record['data']['test'] == rec['test']

    record = db.find('config', id0)
    assert record is None


def test_insert_and_get_current(db):
    rec = {'test': 'insert'}
    db.insert_current('config', rec)

    record = db.get_current('config')
    assert record['data']['test'] == rec['test']


def test_clear_current(db):
    rec = {'test': 'insert'}
    db.insert_current('config', rec)

    record = db.get_current('config')
    assert record['data']['test'] == rec['test']

    db.clear_current('config')

    record = db.get_current('config')
    assert record is None


def test_simple_insert(db):
    rec = {'test': 'insert'}
    # Use `insert` here, which returns an `ObjectId`
    id0 = db.insert('config', rec)

    record = db.find('config', id0)
    assert record['data']['test'] == rec['test']


def test_bad_insert(db):
    """Can't serialize `db` properly so gives warning and returns nothing."""
    with pytest.raises(error.InvalidSerialization):
        rec = db.insert_current('config', db, store_permanently=False)
        assert rec is None

    with pytest.raises(error.InvalidSerialization):
        rec = db.insert('config', db)
        assert rec is None


@pytest.mark.filterwarnings('ignore')
def test_bad_collection(db):
    with pytest.raises(error.InvalidCollection):
        db.insert_current('foobar', {'test': 'insert'})

    with pytest.raises(error.InvalidCollection):
        db.insert('foobar', {'test': 'insert'})


def test_warn_bad_object(db):
    with pytest.raises(error.InvalidSerialization):
        db.insert_current('observations', {'junk': db})

    with pytest.raises(error.InvalidSerialization):
        db.insert('observations', {'junk': db})



import numpy as np

from astropy.stats import SigmaClip

from photutils import Background2D
from photutils import MeanBackground
from photutils import MMMBackground
from photutils import MedianBackground
from photutils import SExtractorBackground
from photutils import BkgZoomInterpolator

from ..logger import logger
from . import fits as fits_utils


def get_rgb_data(data, separate_green=False):
    """Get the data split into separate channels for RGB.

    `data` can be a 2D (`W x H`) or 3D (`N x W x H`) array where `W`=width
    and `H`=height of the data, with `N`=number of frames.

    The return array will be a `3 x W x H` or `3 x N x W x H` array.

    The Bayer array defines a superpixel as a collection of 4 pixels
    set in a square grid:

                     R G
                     G B

    `ds9` and other image viewers define the coordinate axis from the
    lower left corner of the image, which is how a traditional x-y plane
    is defined and how most images would expect to look when viewed. This
    means that the `(0, 0)` coordinate position will be in the lower left
    corner of the image.

    When the data is loaded into a `numpy` array the data is flipped on the
    vertical axis in order to maintain the same indexing/slicing features.
    This means the the `(0, 0)` coordinate position is in the upper-left
    corner of the array when output. When plotting this array one can use
    the `origin='lower'` option to view the array as would be expected in
    a normal image although this does not change the actual index.

    Note:

        Image dimensions:

         ----------------------------
         x | width  | i | columns |  5208
         y | height | j | rows    |  3476

        Bayer Pattern (as seen in ds9):

                                      x / j

                      0     1    2     3 ... 5204 5205 5206 5207
                    --------------------------------------------
               3475 |  R   G1    R    G1        R   G1    R   G1
               3474 | G2    B   G2     B       G2    B   G2    B
               3473 |  R   G1    R    G1        R   G1    R   G1
               3472 | G2    B   G2     B       G2    B   G2    B
                  . |
         y / i    . |
                  . |
                  3 |  R   G1    R    G1        R   G1    R   G1
                  2 | G2    B   G2     B       G2    B   G2    B
                  1 |  R   G1    R    G1        R   G1    R   G1
                  0 | G2    B   G2     B       G2    B   G2    B


        Note: This therefore assumes the data is in the following format:

                 | row (y) |  col (x)
             --------------| ------
              R  |  odd i, |  even j
              G1 |  odd i, |   odd j
              G2 | even i, |  even j
              B  | even i, |   odd j

        Or, in other words, the bottom-left (i.e. `(0,0)`) super-pixel is an RGGB pattern.

        And a mask can therefore be generated as:

            bayer[1::2, 0::2] = 1 # Red
            bayer[1::2, 1::2] = 1 # Green
            bayer[0::2, 0::2] = 1 # Green
            bayer[0::2, 1::2] = 1 # Blue

    """
    rgb_masks = get_rgb_masks(data, separate_green=separate_green)

    color_data = list()

    # Red
    color_data.append(np.ma.array(data, mask=rgb_masks[0]))

    # Green
    color_data.append(np.ma.array(data, mask=rgb_masks[1]))

    if separate_green:
        color_data.append(np.ma.array(data, mask=rgb_masks[2]))

    # Blue
    color_data.append(np.ma.array(data, mask=rgb_masks[-1]))

    return np.ma.array(color_data)


def get_rgb_masks(data, separate_green=False):
    """Get the RGGB Bayer pattern for the given data.

    > Note: See `get_rgb_data` for a description of the RGGB pattern.

    Args:
        data (`numpy.array`): An array of data representing an image.
        separate_green (bool, optional): If the two green channels should be separated,
            default False.

    Returns:
        tuple(np.array, np.array, np.array): A 3-tuple of numpy arrays of `bool` type.
    """

    r_mask = np.ones_like(data).astype(bool)
    g1_mask = np.ones_like(data).astype(bool)
    b_mask = np.ones_like(data).astype(bool)

    if separate_green:
        g2_mask = np.ones_like(data).astype(bool)
    else:
        g2_mask = g1_mask

    if data.ndim == 2:
        r_mask[1::2, 0::2] = False
        g1_mask[1::2, 1::2] = False
        g2_mask[0::2, 0::2] = False
        b_mask[0::2, 1::2] = False
    elif data.ndim == 3:
        r_mask[..., 1::2, 0::2] = False
        g1_mask[..., 1::2, 1::2] = False
        g2_mask[..., 0::2, 0::2] = False
        b_mask[..., 0::2, 1::2] = False
    else:
        raise TypeError('Only 2D and 3D data allowed')

    if separate_green:
        return np.array([r_mask, g1_mask, g2_mask, b_mask])
    else:
        return np.array([r_mask, g1_mask, b_mask])


def get_pixel_color(x, y):
    """ Given an x,y position, return the corresponding color.

    > Note: See `get_rgb_data` for a description of the RGGB pattern.

    Returns:
        str: one of 'R', 'G1', 'G2', 'B'
    """
    x = int(x)
    y = int(y)
    if x % 2 == 0:
        if y % 2 == 0:
            return 'G2'
        else:
            return 'R'
    else:
        if y % 2 == 0:
            return 'B'
        else:
            return 'G1'


def get_rgb_background(fits_fn,
                       box_size=(84, 84),
                       filter_size=(3, 3),
                       camera_bias=0,
                       estimator='mean',
                       interpolator='zoom',
                       sigma=5,
                       iters=5,
                       exclude_percentile=100,
                       return_separate=False,
                       *args,
                       **kwargs
                       ):
    """Get the background for each color channel.

    Most of the options are described in the `photutils.Background2D` page:
    https://photutils.readthedocs.io/en/stable/background.html#d-background-and-noise-estimation

    >>> from panoptes.utils.images import fits as fits_utils
    >>> fits_fn = getfixture('solved_fits_file')

    >>> data = fits_utils.getdata(fits_fn)
    >>> data.mean()
    2236.816...

    >>> rgb_back = get_rgb_background(fits_fn)
    >>> rgb_back.mean()
    2202.392...

    >>> rgb_backs = get_rgb_background(fits_fn, return_separate=True)
    >>> rgb_backs[0]
    <photutils.background.background_2d.Background2D...>
    >>> {color:data.background_rms_median for color, data in zip('rgb', rgb_backs)}
    {'r': 20.566..., 'g': 32.787..., 'b': 23.820...}


    Args:
        fits_fn (str): The filename of the FITS image.
        box_size (tuple, optional): The box size over which to compute the
            2D-Background, default (84, 84).
        filter_size (tuple, optional): The filter size for determining the median,
            default (3, 3).
        camera_bias (int, optional): The built-in camera bias, default 0. A zero camera
            bias means the bias will be considered as part of the background.
        estimator (str, optional): The estimator object to use, default 'median'.
        interpolator (str, optional): The interpolater object to user, default 'zoom'.
        sigma (int, optional): The sigma on which to filter values, default 5.
        iters (int, optional): The number of iterations to sigma filter, default 5.
        exclude_percentile (int, optional): The percentage of the data (per channel)
            that can be masked, default 100 (i.e. all).
        return_separate (bool, optional): If the function should return a separate array
            for color channel, default False.
        *args: Description
        **kwargs: Description

    Returns:
        `numpy.array`|list: Either a single numpy array representing the entire
          background, or a list of masked numpy arrays in RGB order. The background
          for each channel has full interploation across all pixels, but the mask covers
          them.
    """
    logger.info(f"Getting background for {fits_fn}")
    logger.debug(
        f"{estimator} {interpolator} {box_size} {filter_size} {camera_bias} σ={sigma} n={iters}")

    estimators = {
        'sexb': SExtractorBackground,
        'median': MedianBackground,
        'mean': MeanBackground,
        'mmm': MMMBackground
    }
    interpolators = {
        'zoom': BkgZoomInterpolator,
    }

    bkg_estimator = estimators[estimator]()
    interp = interpolators[interpolator]()

    data = fits_utils.getdata(fits_fn) - camera_bias

    # Get the data per color channel.
    rgb_data = get_rgb_data(data)

    backgrounds = list()
    for color, color_data in zip(['R', 'G', 'B'], rgb_data):
        logger.debug(f'Performing background {color} for {fits_fn}')

        bkg = Background2D(color_data,
                           box_size,
                           filter_size=filter_size,
                           sigma_clip=SigmaClip(sigma=sigma, maxiters=iters),
                           bkg_estimator=bkg_estimator,
                           exclude_percentile=exclude_percentile,
                           mask=color_data.mask,
                           interpolator=interp)

        # Create a masked array for the background
        if return_separate:
            backgrounds.append(bkg)
        else:
            backgrounds.append(np.ma.array(data=bkg.background, mask=color_data.mask))
        logger.debug(
            f"{color} Value: {bkg.background_median:.02f} RMS: {bkg.background_rms_median:.02f}")

    if return_separate:
        return backgrounds

    # Create one array for the backgrounds, where any holes are filled with zeros.
    full_background = np.ma.array(backgrounds).sum(0).filled(0)

    return full_background



import pytest
import numpy as np

from panoptes.utils.images import bayer


def test_get_rgb_2d_data():
    data_2d = np.ones((10, 10))
    rgb_data = bayer.get_rgb_data(data_2d)

    assert isinstance(rgb_data, np.ma.core.MaskedArray)
    assert len(rgb_data) == 3

    # Red and Blue will add to 25 for a 10x10 data stamp
    assert rgb_data[0].sum() == 25
    assert rgb_data[-1].sum() == 25

    assert rgb_data[1].sum() == 50

    rgb_data = bayer.get_rgb_data(data_2d, separate_green=True)

    assert isinstance(rgb_data, np.ma.core.MaskedArray)
    assert len(rgb_data) == 4

    # Red and Blue will add to 25 for a 10x10 data stamp
    assert rgb_data[0].sum() == 25
    assert rgb_data[1].sum() == 25
    assert rgb_data[2].sum() == 25
    assert rgb_data[3].sum() == 25


def test_get_rgb_3d_data():
    data_2d = np.ones((10, 10, 10))
    rgb_data = bayer.get_rgb_data(data_2d)

    assert isinstance(rgb_data, np.ma.core.MaskedArray)
    assert len(rgb_data) == 3

    # Red and Blue will add to 25 for a 10x10 data stamp
    assert rgb_data[0].sum() == 250
    assert rgb_data[-1].sum() == 250

    assert rgb_data[1].sum() == 500

    rgb_data = bayer.get_rgb_data(data_2d, separate_green=True)

    assert isinstance(rgb_data, np.ma.core.MaskedArray)
    assert len(rgb_data) == 4

    # Red and Blue will add to 25 for a 10x10 data stamp
    assert rgb_data[0].sum() == 250
    assert rgb_data[1].sum() == 250
    assert rgb_data[2].sum() == 250
    assert rgb_data[3].sum() == 250


def test_get_rgb_4d_data():
    data = np.ones((10, 10, 10, 10))
    with pytest.raises(TypeError):
        bayer.get_rgb_data(data)


def test_get_pixel_color():
    """
        From the docstring:

                 | row (y) |  col (x)
             --------------| ------
              R  |  odd i, |  even j
              G1 |  odd i, |   odd j
              G2 | even i, |  even j
              B  | even i, |   odd j

    """

    assert bayer.get_pixel_color(0, 1) == 'R'
    assert bayer.get_pixel_color(1, 1) == 'G1'
    assert bayer.get_pixel_color(2, 2) == 'G2'
    assert bayer.get_pixel_color(1, 2) == 'B'

    # Test with some fractional pixels
    assert bayer.get_pixel_color(0, 1.1) == 'R'
    assert bayer.get_pixel_color(1.9, 1) == 'G1'
    assert bayer.get_pixel_color(2, 2.5) == 'G2'
    assert bayer.get_pixel_color(1.5, 2) == 'B'



import pytest
import tweepy
import requests
import unittest.mock

from panoptes.utils.social.twitter import SocialTwitter
from panoptes.utils.social.slack import SocialSlack


@pytest.fixture(scope='module')
def twitter_config():
    twitter_config = {'consumer_key': 'mock_consumer_key', 'consumer_secret': 'mock_consumer_secret',
                      'access_token': 'mock_access_token', 'access_token_secret': 'access_token_secret'}
    return twitter_config


@pytest.fixture(scope='module')
def slack_config():
    slack_config = {'webhook_url': 'mock_webhook_url', 'output_timestamp': False}
    return slack_config


# Twitter sink tests
def test_no_consumer_key(twitter_config):
    with unittest.mock.patch.dict(twitter_config), pytest.raises(ValueError) as ve:
        del twitter_config['consumer_key']
        SocialTwitter(**twitter_config)
        assert False  # We don't reach this point
    assert 'consumer_key parameter is not defined.' == str(ve.value)


def test_no_consumer_secret(twitter_config):
    with unittest.mock.patch.dict(twitter_config), pytest.raises(ValueError) as ve:
        del twitter_config['consumer_secret']
        SocialTwitter(**twitter_config)
        assert False  # We don't reach this point
    assert 'consumer_secret parameter is not defined.' == str(ve.value)


def test_no_access_token(twitter_config):
    with unittest.mock.patch.dict(twitter_config), pytest.raises(ValueError) as ve:
        del twitter_config['access_token']
        SocialTwitter(**twitter_config)
        assert False  # We don't reach this point
    assert 'access_token parameter is not defined.' == str(ve.value)


def test_no_access_token_secret(twitter_config):
    with unittest.mock.patch.dict(twitter_config), pytest.raises(ValueError) as ve:
        del twitter_config['access_token_secret']
        SocialTwitter(**twitter_config)
        assert False  # We don't reach this point
    assert 'access_token_secret parameter is not defined.' == str(ve.value)


def test_send_message_twitter(twitter_config):
    with unittest.mock.patch.object(tweepy.API, 'update_status') as mock_update_status:
        social_twitter = SocialTwitter(**twitter_config)
        mock_message = "mock_message"
        mock_timestamp = "mock_timestamp"
        social_twitter.send_message(mock_message, mock_timestamp)

        mock_update_status.assert_called_once_with('{} - {}'.format(mock_message, mock_timestamp))


def test_send_message_twitter_no_timestamp(twitter_config):
    with unittest.mock.patch.dict(twitter_config, {'output_timestamp': False}), unittest.mock.patch.object(tweepy.API, 'update_status') as mock_update_status:
        social_twitter = SocialTwitter(**twitter_config)
        mock_message = "mock_message"
        mock_timestamp = "mock_timestamp"
        social_twitter.send_message(mock_message, mock_timestamp)

        mock_update_status.assert_called_once_with(mock_message)


# Slack sink tests
def test_no_webhook_url(slack_config):
    with unittest.mock.patch.dict(slack_config), pytest.raises(ValueError) as ve:
        del slack_config['webhook_url']
        slack_config = SocialSlack(**slack_config)
    assert 'webhook_url parameter is not defined.' == str(ve.value)


def test_send_message_slack(slack_config):
    with unittest.mock.patch.object(requests, 'post') as mock_post:
        social_slack = SocialSlack(**slack_config)
        mock_message = "mock_message"
        mock_timestamp = "mock_timestamp"
        social_slack.send_message(mock_message, mock_timestamp)

        mock_post.assert_called_once_with(slack_config['webhook_url'], json={'text': mock_message})


def test_send_message_slack_timestamp(slack_config):
    with unittest.mock.patch.dict(slack_config, {'output_timestamp': True}), unittest.mock.patch.object(requests, 'post') as mock_post:
        social_slack = SocialSlack(**slack_config)
        mock_message = "mock_message"
        mock_timestamp = "mock_timestamp"
        social_slack.send_message(mock_message, mock_timestamp)

        mock_post.assert_called_once_with(slack_config['webhook_url'], json={
                                          'text': '{} - {}'.format(mock_message, mock_timestamp)})



# This module implements a handler for serial_for_url("buffers://").

from panoptes.utils.tests.serial_handlers import NoOpSerial

import io
from serial import serialutil
import threading

# r_buffer and w_buffer are binary I/O buffers. read(size=N) on an instance
# of Serial reads the next N bytes from r_buffer, and write(data) appends the
# bytes of data to w_buffer.
# NOTE: The caller (a test) is responsible for resetting buffers before tests.
_r_buffer = None
_w_buffer = None

# The above I/O buffers are not thread safe, so we need to lock them during
# access.
_r_lock = threading.Lock()
_w_lock = threading.Lock()


def ResetBuffers(read_data=None):
    SetRBufferValue(read_data)
    with _w_lock:
        global _w_buffer
        _w_buffer = io.BytesIO()


def SetRBufferValue(data):
    """Sets the r buffer to data (a bytes object)."""
    if data and not isinstance(data, (bytes, bytearray)):
        raise TypeError("data must by a bytes or bytearray object.")
    with _r_lock:
        global _r_buffer
        _r_buffer = io.BytesIO(data)


def GetWBufferValue():
    """Returns an immutable bytes object with the value of the w buffer."""
    with _w_lock:
        if _w_buffer:
            return _w_buffer.getvalue()


class BuffersSerial(NoOpSerial):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def in_waiting(self):
        if not self.is_open:
            raise serialutil.portNotOpenError
        with _r_lock:
            return len(_r_buffer.getbuffer()) - _r_buffer.tell()

    def read(self, size=1):
        """Read size bytes.

        If a timeout is set it may return fewer characters than requested.
        With no timeout it will block until the requested number of bytes
        is read.

        Args:
            size: Number of bytes to read.

        Returns:
            Bytes read from the port, of type 'bytes'.

        Raises:
            SerialTimeoutException: In case a write timeout is configured for
                the port and the time is exceeded.
        """
        if not self.is_open:
            raise serialutil.portNotOpenError
        with _r_lock:
            # TODO(jamessynge): Figure out whether and how to handle timeout.
            # We might choose to generate a timeout if the caller asks for data
            # beyond the end of the buffer; or simply return what is left,
            # including nothing (i.e. bytes()) if there is nothing left.
            return _r_buffer.read(size)

    def write(self, data):
        """
        Args:
            data: The data to write.

        Returns:
            Number of bytes written.

        Raises:
            SerialTimeoutException: In case a write timeout is configured for
                the port and the time is exceeded.
        """
        if not isinstance(data, (bytes, bytearray)):
            raise TypeError("data must by a bytes or bytearray object.")
        if not self.is_open:
            raise serialutil.portNotOpenError
        with _w_lock:
            return _w_buffer.write(data)


Serial = BuffersSerial



import os
from warnings import warn
from contextlib import suppress
from uuid import uuid4
from glob import glob

from .. import error
from ..serializers import to_json
from ..serializers import from_json
from ..database import AbstractPanDB
from ..database import create_storage_obj


class PanFileDB(AbstractPanDB):
    """Stores collections as files of JSON records."""

    def __init__(self, db_name='panoptes', **kwargs):
        """Flat file storage for json records

        This will simply store each json record inside a file corresponding
        to the type. Each entry will be stored in a single line.
        Args:
            db_name (str, optional): Name of the database containing the collections.
        """

        super().__init__(db_name=db_name, **kwargs)

        self.db_folder = db_name

        # Set up storage directory.
        self._storage_dir = os.path.join(os.environ['PANDIR'], 'json_store', self.db_folder)
        os.makedirs(self._storage_dir, exist_ok=True)

    def insert_current(self, collection, obj, store_permanently=True):
        self.validate_collection(collection)
        obj_id = self._make_id()
        obj = create_storage_obj(collection, obj, obj_id=obj_id)
        current_fn = self._get_file(collection, permanent=False)
        result = obj_id

        try:
            # Overwrite current collection file with obj.
            to_json(obj, filename=current_fn, append=False)
        except Exception as e:
            raise error.InvalidSerialization(f"Problem serializing object for insertion: {e} {current_fn} {obj!r}")

        if not store_permanently:
            return result
        else:
            return self.insert(collection, obj)

    def insert(self, collection, obj):
        self.validate_collection(collection)
        obj_id = self._make_id()
        obj = create_storage_obj(collection, obj, obj_id=obj_id)
        collection_fn = self._get_file(collection)
        try:
            # Insert record into file
            to_json(obj, filename=collection_fn)
            return obj_id
        except Exception as e:
            raise error.InvalidSerialization(f"Problem inserting object into collection: {e}, {obj!r}")

    def get_current(self, collection):
        current_fn = self._get_file(collection, permanent=False)

        try:
            with open(current_fn) as f:
                msg = from_json(f.read())

            return msg
        except FileNotFoundError:
            self.logger.warning("No record found for {}".format(collection))
            return None

    def find(self, collection, obj_id):
        collection_fn = self._get_file(collection)
        obj = None
        with suppress(FileNotFoundError):
            with open(collection_fn, 'r') as f:
                for line in f:
                    if obj_id in line:
                        obj = from_json(line)
                        break

        return obj

    def clear_current(self, record_type):
        """Clears the current record of the given type.

        Args:
            record_type (str): The record type, e.g. 'weather', 'environment', etc.
        """
        current_f = self._get_file(record_type, permanent=False)
        with suppress(FileNotFoundError):
            os.remove(current_f)

    def _get_file(self, collection, permanent=True):
        if permanent:
            name = '{}.json'.format(collection)
        else:
            name = 'current_{}.json'.format(collection)
        return os.path.join(self._storage_dir, name)

    def _make_id(self):
        return str(uuid4())

    @classmethod
    def permanently_erase_database(cls, db_name):
        # Clear out any .json files.
        storage_dir = os.path.join(os.environ['PANDIR'], 'json_store', db_name)
        for f in glob(os.path.join(storage_dir, '*.json')):
            os.remove(f)



import abc

from ..logger import logger
from .. import error
from ..time import current_time
from ..library import load_module


def _get_db_class(module_name='file'):
    """Load the main DB class for the module of the given name.

    .. doctest:

        >>> _get_db_class()
        <class 'panoptes.utils.database.file.PanFileDB'>
        >>> _get_db_class('memory')
        <class 'panoptes.utils.database.memory.PanMemoryDB'>

    Args:
        module_name (str): Name of module, one of: `file` (default), 'memory'.

    Returns:
        `panoptes.utils.database.PanDB`: An instance of the db class for the correct database type.

    Raises:
        Exception: If an unsupported database type string is passed.
    """
    class_map = {
        'file': 'PanFileDB',
        'memory': 'PanMemoryDB',
    }

    full_module_name = f'panoptes.utils.database.{module_name}'

    try:
        db_module = load_module(full_module_name)
        return getattr(db_module, class_map[module_name])
    except Exception:
        raise Exception(f'Unsupported database type: {full_module_name}')


class AbstractPanDB(metaclass=abc.ABCMeta):
    def __init__(self, db_name=None, collection_names=list(), **kwargs):
        """
        Init base class for db instances.

        Args:
            db_name: Name of the database, typically panoptes or panoptes_testing.
            collection_names (list of str): Names of the valid collections.
            logger: (Optional) logger to use for warnings.
        """
        self.logger = logger
        self.logger.info(f'Creating PanDB {db_name} with collections: {collection_names}')
        self.db_name = db_name
        self.collection_names = collection_names

    def validate_collection(self, collection):
        if collection not in self.collection_names:
            raise error.InvalidCollection(f'Collection type {collection!r} not available')

    @abc.abstractclassmethod
    def insert_current(self, collection, obj, store_permanently=True):  # pragma: no cover
        """Insert an object into both the `current` collection and the collection provided.

        Args:
            collection (str): Name of valid collection within the db.
            obj (dict or str): Object to be inserted.
            store_permanently (bool): Whether to also update the collection,
                defaults to True.

        Returns:
            str: identifier of inserted record. If `store_permanently` is True, will
                be the identifier of the object in the `collection`, otherwise will be the
                identifier of object in the `current` collection. These may or
                may not be the same.
                Returns None if unable to insert into the collection.
        """
        raise NotImplementedError

    @abc.abstractclassmethod
    def insert(self, collection, obj):  # pragma: no cover
        """Insert an object into the collection provided.

        The `obj` to be stored in a collection should include the `type`
        and `date` metadata as well as a `data` key that contains the actual
        object data. If these keys are not provided then `obj` will be wrapped
        in a corresponding object that does contain the metadata.

        Args:
            collection (str): Name of valid collection within the db.
            obj (dict or str): Object to be inserted.

        Returns:
            str: identifier of inserted record in `collection`.
                Returns None if unable to insert into the collection.
        """
        raise NotImplementedError

    @abc.abstractclassmethod
    def get_current(self, collection):  # pragma: no cover
        """Returns the most current record for the given collection

        Args:
            collection (str): Name of the collection to get most current from

        Returns:
            dict|None: Current object of the collection or None.
        """
        raise NotImplementedError

    @abc.abstractclassmethod
    def find(self, collection, obj_id):  # pragma: no cover
        """Find an object by it's identifier.

        Args:
            collection (str): Collection to search for object.
            obj_id (ObjectID|str): Record identifier returned earlier by insert
                or insert_current.

        Returns:
            dict|None: Object matching identifier or None.
        """
        raise NotImplementedError

    @abc.abstractclassmethod
    def clear_current(self, type):  # pragma: no cover
        """Clear the current record of a certain type

        Args:
            type (str): The type of entry in the current collection that
                should be cleared.
        """
        raise NotImplementedError


def create_storage_obj(collection, data, obj_id=None):
    """Returns the object to be stored in the database"""
    obj = dict(data=data, type=collection, date=current_time(datetime=True))
    if obj_id:
        obj['_id'] = obj_id
    return obj


class PanDB(object):
    """Simple class to load the appropriate DB type based on the config.

    We don't actually create instances of this class, but instead create
    an instance of the 'correct' type of db.
    """

    def __new__(cls, db_type='memory', db_name=None, *args, **kwargs):
        """Create an instance based on db_type."""

        collection_names = PanDB.collection_names()

        # Load the correct DB module
        DB = _get_db_class(db_type)

        if db_type == 'memory':
            # The memory type has special setup
            db_instance = DB.get_or_create(collection_names=collection_names, **kwargs)
        else:
            db_instance = DB(collection_names=collection_names, **kwargs)

        return db_instance

    @staticmethod
    def collection_names():
        """The pre-defined list of collections that are valid."""
        return [
            'camera_board',
            'control_board',
            'camera_env_board',
            'control_env_board',
            'config',
            'current',
            'drift_align',
            'environment',
            'mount',
            'observations',
            'offset_info',
            'power',
            'safety',
            'state',
            'telemetry_board',
            'weather',
        ]

    @classmethod
    def permanently_erase_database(cls,
                                   db_type,
                                   db_name,
                                   really=False,
                                   dangerous=False,
                                   *args, **kwargs):
        """Permanently delete the contents of the identified database."""

        if not isinstance(db_name, str) or 'test' not in db_name:
            raise ValueError(
                'permanently_erase_database() called for non-test database {!r}'.format(db_name))

        if really != 'Yes' or dangerous != 'Totally':
            raise Exception('PanDB.permanently_erase_database called with invalid args!')

        # Load the correct DB module.
        DB = _get_db_class(db_type)

        # Do the deletion.
        DB.permanently_erase_database(db_name, *args, **kwargs)



import multiprocessing
import pytest
import time

from datetime import datetime
from panoptes.utils.messaging import PanMessaging


@pytest.fixture(scope='module')
def mp_manager():
    return multiprocessing.Manager()


@pytest.fixture(scope='function')
def forwarder(mp_manager):
    ready = mp_manager.Event()
    done = mp_manager.Event()

    def start_forwarder():
        PanMessaging.create_forwarder(
            12345, 54321, ready_fn=lambda: ready.set(), done_fn=lambda: done.set())

    messaging = multiprocessing.Process(target=start_forwarder)
    messaging.start()

    if not ready.wait(timeout=10.0):
        raise Exception('Forwarder failed to become ready!')
    # Wait a moment for the forwarder to start using those sockets.
    time.sleep(0.05)

    yield messaging

    # Stop the forwarder. Since we use the same ports in multiple
    # tests, we wait for the process to shutdown.
    messaging.terminate()
    for _ in range(100):
        # We can't be sure that the sub-process will succeed in
        # calling the done_fn, so we also check for the process
        # ending.
        if done.wait(timeout=0.01):
            break
        if not messaging.is_alive():
            break


def test_forwarder(forwarder):
    assert forwarder.is_alive() is True


@pytest.fixture(scope='function')
def pub_and_sub(forwarder):
    # Ensure that the subscriber is created first.
    sub = PanMessaging.create_subscriber(54321)
    time.sleep(0.05)
    pub = PanMessaging.create_publisher(12345, bind=False, connect=True)
    time.sleep(0.05)
    yield (pub, sub)
    pub.close()
    sub.close()


def test_send_string(pub_and_sub):
    pub, sub = pub_and_sub
    pub.send_message('Test-Topic', 'Hello')
    topic, msg_obj = sub.receive_message()

    assert topic == 'Test-Topic'
    assert isinstance(msg_obj, dict)
    assert 'message' in msg_obj
    assert msg_obj['message'] == 'Hello'


def test_send_datetime(pub_and_sub):
    pub, sub = pub_and_sub
    date_obj = datetime(2017, 1, 1)
    pub.send_message('Test-Topic', {'date': date_obj})
    topic, msg_obj = sub.receive_message()
    assert msg_obj['date'] == date_obj


################################################################################
# Tests of the conftest.py messaging fixtures.

def test_message_forwarder_exists(message_forwarder):
    assert isinstance(message_forwarder, dict)
    assert 'msg_ports' in message_forwarder

    assert isinstance(message_forwarder['msg_ports'], tuple)
    assert len(message_forwarder['msg_ports']) == 2
    assert isinstance(message_forwarder['msg_ports'][0], int)
    assert isinstance(message_forwarder['msg_ports'][1], int)

    assert isinstance(message_forwarder['cmd_ports'], tuple)
    assert len(message_forwarder['cmd_ports']) == 2
    assert isinstance(message_forwarder['cmd_ports'][0], int)
    assert isinstance(message_forwarder['cmd_ports'][1], int)

    # The ports should be unique.
    msg_ports = message_forwarder['msg_ports']
    cmd_ports = message_forwarder['cmd_ports']

    ports = set(list(msg_ports) + list(cmd_ports))
    assert len(ports) == 4


def assess_pub_sub(pub, sub):
    """Helper method for testing a pub-sub pair."""

    # Can not send a message using a subscriber
    print("Can't send using subscriber")
    with pytest.raises(Exception):
        sub.send_message('topic_name', 'a string')

    # Can not receive a message using a publisher
    print("Can't receive with publisher")
    assert (None, None) == pub.receive_message(blocking=True, timeout_ms=1000)

    # At first, there is nothing available to receive.
    print("Nothing at start")
    assert (None, None) == sub.receive_message(blocking=True, timeout_ms=500)

    print("Sending a string")
    pub.send_message('topic.name', 'a string')

    print("Receiving a string")
    topic, msg_obj = sub.receive_message(timeout_ms=5000)

    print("Checking string")
    assert isinstance(msg_obj, dict)
    assert 'message' in msg_obj
    assert msg_obj['message'] == 'a string'
    assert 'timestamp' in msg_obj


def test_msg_pub_sub(message_forwarder, msg_publisher, msg_subscriber):
    print("Calling helper")
    print(f"Using messaging_ports: {message_forwarder}")
    assess_pub_sub(msg_publisher, msg_subscriber)


def test_cmd_pub_sub(message_forwarder, cmd_publisher, cmd_subscriber):
    assess_pub_sub(cmd_publisher, cmd_subscriber)



import contextlib
import os
import re
import shutil
import signal
import collections.abc
from urllib.parse import unquote as unquote_url

import numpy as np
from astropy import units as u
from astropy.coordinates import AltAz
from astropy.coordinates import ICRS
from astropy.coordinates import SkyCoord

from .time import current_time


PATH_MATCHER = re.compile(r'''
    .*?
    (?P<unit_id>PAN\d{3})[/_]{1}
    (?P<camera_id>[a-gA-G0-9]{6})[/_]{1}
    (?P<sequence_id>[0-9]{8}T[0-9]{6})[/_]{1}
    (?P<image_id>[0-9]{8}T[0-9]{6})
    .*?
''', re.VERBOSE)


def listify(obj):
    """ Given an object, return a list.

    Always returns a list. If obj is None, returns empty list,
    if obj is list, just returns obj, otherwise returns list with
    obj as single member.

    If a `dict` object is passed then this function will return a list of *only*
    the values.

    .. doctest::

        >>> listify(42)
        [42]
        >>> listify('foo')
        ['foo']
        >>> listify(None)
        []
        >>> listify(['a'])
        ['a']

        >>> my_dict = dict(a=42, b='foo')
        >>> listify(my_dict)
        [42, 'foo']
        >>> listify(my_dict.values())
        [42, 'foo']
        >>> listify(my_dict.keys())
        ['a', 'b']


    Returns:
        list:   You guessed it.
    """
    if obj is None:
        return list()
    elif isinstance(obj, list):
        return obj
    elif isinstance(obj, dict):
        return list(obj.values())
    elif isinstance(obj, (collections.abc.ValuesView, collections.abc.KeysView)):
        return list(obj)
    else:
        return [obj]


def get_free_space(dir=None):
    """Return the amoung of freespace in gigabytes for given dir.

    >>> from panoptes.utils import get_free_space
    >>> get_free_space()
    <Quantity ... Gbyte>

    >>> get_free_space(dir='/')
    <Quantity ... Gbyte>

    Args:
        dir (str, optional): Path to directory. If None defaults to $PANDIR.

    Returns:
        astropy.units.Quantity: The number of gigabytes avialable in folder.

    """
    if dir is None:
        dir = os.getenv('PANDIR')

    _, _, free_space = shutil.disk_usage(dir)
    free_space = (free_space * u.byte).to(u.gigabyte)
    return free_space


def string_to_params(opts):
    """Parses a single string into parameters that can be passed to a function.

    A user of the `peas_shell` can supply positional and keyword arguments to the
    command being called, however the `Cmd` module that is used for the shell does
    not parse these options but instead passes this as a single string. This utility
    method does some simple parsing of that string and returns a list of positional
    parameters and a dictionary of keyword arguments.  A keyword argument is considered
    anything that contains an equal sign (e.g. `exptime=30`). Any leading `--` to
    a keyword argument will be stripped during parsing.

    A list of items can be passed by specifying the keyword argument multiple times.

    Note:

        This function will attempt to parse keyword values as floats if possible.
        If a string is required include a single quote around the value, e.g.
        `param='42'` will keep the value as the string `'42'`.


    >>> from panoptes.utils import string_to_params
    >>> args, kwargs = string_to_params("parg1 parg2 key1=a_str key2=2 key2='2' key3=03")
    >>> args
    ['parg1', 'parg2']
    >>> kwargs
    {'key1': 'a_str', 'key2': [2.0, '2'], 'key3': 3.0}
    >>> isinstance(kwargs['key2'][0], float)
    True
    >>> isinstance(kwargs['key2'][1], str)
    True
    >>> kwargs['key2'][1] == '2'
    True

    >>> args, kwargs = string_to_params('--key1=val1 --key1-2=val1-2')
    >>> kwargs
    {'key1': 'val1', 'key1-2': 'val1-2'}

    Args:
        opts (str): A single string containing everything beyond the actual
            command that is called.

    Returns:
        tuple(list, dict): Returns a list of positional parameters and a dictionary
        of keyword arguments. These correspond to the *args and **kwargs that
        a typical function would receive.
    """
    args = []
    kwargs = {}

    for opt in opts.split(' '):
        if '=' not in opt:
            args.append(opt)
        else:
            name, value = opt.split('=', maxsplit=1)
            if name.startswith('--') and len(name) > 2:
                name = name[2:]

            if "'" in value:
                # Remove the explict single quotes.
                value = value.replace("'", "")
            else:
                # Make it a number if possible.
                with contextlib.suppress(ValueError):
                    value = float(value)

            if name in kwargs:
                kwargs[name] = listify(kwargs[name])
                kwargs[name].append(value)
            else:
                kwargs[name] = value

    return args, kwargs


def altaz_to_radec(alt=35, az=90, location=None, obstime=None, **kwargs):
    """Convert alt/az degrees to RA/Dec SkyCoord.

    >>> from panoptes.utils import altaz_to_radec
    >>> from astropy.coordinates import EarthLocation
    >>> keck = EarthLocation.of_site('Keck Observatory')
    ...

    >>> altaz_to_radec(alt=75, az=180, location=keck, obstime='2020-02-02T20:20:02.02')
    <SkyCoord (ICRS): (ra, dec) in deg
        (281.78..., 4.807...)>

    >>> # Will use current time if none given
    >>> altaz_to_radec(location=keck)
    <SkyCoord (ICRS): (ra, dec) in deg
        (..., ...)>

    >>> altaz_to_radec(location=keck, obstime='2020-02-02T20:20:02.02')
    <SkyCoord (ICRS): (ra, dec) in deg
        (338.4096..., 11.1175...)>

    >>> # Must pass a `location` instance.
    >>> altaz_to_radec()
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
      ...
        assert location is not None
    AssertionError

    Args:
        alt (int, optional): Altitude, defaults to 35
        az (int, optional): Azimute, defaults to 90 (east)
        location (None|astropy.coordinates.EarthLocation, required): A valid location.
        obstime (None, optional): Time for object, defaults to `current_time`

    Returns:
        astropy.coordinates.SkyCoord: Coordinates corresponding to the AltAz.
    """
    assert location is not None
    if obstime is None:
        obstime = current_time()

    altaz = AltAz(obstime=obstime, location=location, alt=alt * u.deg, az=az * u.deg)
    return SkyCoord(altaz.transform_to(ICRS))


class DelaySigTerm(contextlib.ContextDecorator):
    """Supports delaying SIGTERM during a critical section.

    This allows one to avoid having SIGTERM interrupt a
    critical block of code, such as saving to a database.

    Example:

        ..
        with DelaySigTerm():
            db.WriteCurrentRecord(record)

    """
    # TODO(jamessynge): Consider generalizing as DelaySignal(signum).

    def __enter__(self, callback=None):
        """
        Args:
            callback: If not None, called when SIGTERM is handled,
                with kwargs previously_caught and frame.
        """
        self.caught = False
        self.old_handler = signal.getsignal(signal.SIGTERM)
        if callback:
            assert callable(callback)
            self.callback = callback
        else:
            self.callback = None

        def handler(signum, frame):
            previously_caught = self.caught
            self.caught = True
            if self.callback:
                self.callback(previously_caught=previously_caught, frame=frame)

        signal.signal(signal.SIGTERM, handler)
        return self

    def __exit__(self, *exc):
        signal.signal(signal.SIGTERM, self.old_handler)
        if self.caught:
            # Send SIGTERM to this process.
            os.kill(os.getpid(), signal.SIGTERM)
            # Suppress any exception caught while the context was running.
            return True
        return False


def get_quantity_value(quantity, unit=None):
    """ Thin-wrapper around the `astropy.units.Quantity.to_value` method.

    If passed something other than a Quantity will simply return the original object.

    >>> from astropy import units as u
    >>> from panoptes.utils import get_quantity_value

    >>> get_quantity_value(60 * u.second)
    60.0

    >>> # Can convert between units.
    >>> get_quantity_value(60 * u.minute, unit='second')
    3600.0

    >>> get_quantity_value(60 * u.minute, unit=u.second)
    3600.0

    >>> get_quantity_value(60)
    60

    Args:
        quantity (astropy.units.Quantity): Quantity to extract numerical value from.
        unit (astropy.units.Unit, optional): unit to convert to.

    Returns:
        float: numerical value of the Quantity after conversion to the specified unit.
    """
    try:
        return quantity.to_value(unit)
    except AttributeError:
        return quantity


def moving_average(data_set, periods=3):
    """Moving average.

    Args:
        data_set (`numpy.array`): An array of values over which to perform the moving average.
        periods (int, optional): Number of periods.

    Returns:
        `numpy.array`: An array of the computed averages.
    """
    weights = np.ones(periods) / periods
    return np.convolve(data_set, weights, mode='same')


def image_id_from_path(path):
    """Return the `image_id` from the given path or uri.

    >>> from panoptes.utils import image_id_from_path
    >>> path = 'gs://panoptes-raw-images/PAN012/ee04d1/20190820T111638/20190820T122447.fits'
    >>> image_id_from_path(path)
    'PAN012_ee04d1_20190820T122447'

    >>> path = 'nothing/to/match'
    >>> image_id_from_path(path)


    Args:
        path (str): A path or uri for a file.

    Returns:
        str: The image id in the form "<unit_id>_<camera_id>_<image_id>"
    """
    result = PATH_MATCHER.match(unquote_url(path))

    with contextlib.suppress(AttributeError):
        match = '{}_{}_{}'.format(
            result.group('unit_id'),
            result.group('camera_id'),
            result.group('image_id')
        )
        return match

    return None


def sequence_id_from_path(path):
    """Return the `sequence_id` from the given path or uri.

    >>> from panoptes.utils import sequence_id_from_path
    >>> path = 'gs://panoptes-raw-images/PAN012/ee04d1/20190820T111638/20190820T122447.fits'
    >>> sequence_id_from_path(path)
    'PAN012_ee04d1_20190820T111638'

    >>> path = 'nothing/to/match'
    >>> sequence_id_from_path(path)


    Args:
        path (str): A path or uri for a file.

    Returns:
        str: The image id in the form "<unit_id>_<camera_id>_<sequence_id>"
    """
    result = PATH_MATCHER.match(unquote_url(path))

    with contextlib.suppress(AttributeError):
        match = '{}_{}_{}'.format(
            result.group('unit_id'),
            result.group('camera_id'),
            result.group('sequence_id')
        )
        return match

    return None



import requests

from ..logger import logger
from ..serializers import from_json
from ..serializers import to_json


def get_config(key=None, host='localhost', port='6563', parse=True, default=None):
    """Get a config item from the config server.

    Return the config entry for the given `key`. If `key=None` (default), return
    the entire config.

    Nested keys can be specified as a string, as per [scalpl](https://pypi.org/project/scalpl/).

    Examples:
        >>> get_config(key='name')
        'Generic PANOPTES Unit'
        >>> get_config(key='location.horizon')
        <Quantity 30. deg>
        >>> get_config(key='location.horizon', parse=False)
        '30.0 deg'
        >>> get_config(key='cameras.devices[1].model')
        'canon_gphoto2'
        >>> # Returns `None` if key is not found.
        >>> foobar = get_config(key='foobar')
        >>> foobar is None
        True
        >>> # But you can supply a default.
        >>> get_config(key='foobar', default='baz')
        'baz'
        >>> # Can use Quantities as well
        >>> from astropy import units as u
        >>> get_config(key='foobar', default=42 * u.meter)
        <Quantity 42. m>

    Args:
        key (str): The key to update, see Examples in `get_config` for details.
        host (str, optional): The config server host, defaults to '127.0.0.1'.
        port (str, optional): The config server port, defaults to 6563.
        parse (bool, optional): If response should be parsed by
            `~panoptes.utils.serializers.from_json`, default True.
        default (str, optional): The config server port, defaults to 6563.

    Returns:
        dict: The corresponding config entry.

    Raises:
        Exception: Raised if the config server is not available.
    """
    url = f'http://{host}:{port}/get-config'

    config_entry = default

    try:
        response = requests.post(url, json={'key': key})
    except Exception as e:
        logger.info(f'Problem with get_config: {e!r}')
    else:
        if not response.ok:
            logger.info(f'Problem with get_config: {response.content!r}')
        else:
            if response.text != 'null\n':
                if parse:
                    config_entry = from_json(response.content.decode('utf8'))
                else:
                    config_entry = response.json()

    if config_entry is None:
        config_entry = default

    return config_entry


def set_config(key, new_value, host='localhost', port='6563', parse=True):
    """Set config item in config server.

    Given a `key` entry, update the config to match. The `key` is a dot accessible
    string, as given by [scalpl](https://pypi.org/project/scalpl/). See Examples in `get_config`
    for details.

    Examples:
        >>> from astropy import units as u
        >>> set_config('location.horizon', 35 * u.degree)
        {'location.horizon': <Quantity 35. deg>}
        >>> get_config(key='location.horizon')
        <Quantity 35. deg>
        >>> set_config('location.horizon', 30 * u.degree)
        {'location.horizon': <Quantity 30. deg>}

    Args:
        key (str): The key to update, see Examples in `get_config` for details.
        new_value (scalar|object): The new value for the key, can be any serializable object.
        host (str, optional): The config server host, defaults to '127.0.0.1'.
        port (str, optional): The config server port, defaults to 6563.
        parse (bool, optional): If response should be parsed by
            `~panoptes.utils.serializers.from_json`, default True.

    Returns:
        dict: The updated config entry.

    Raises:
        Exception: Raised if the config server is not available.
    """
    url = f'http://{host}:{port}/set-config'

    json_str = to_json({key: new_value})

    config_entry = None
    try:
        # We use our own serializer so pass as `data` instead of `json`.
        response = requests.post(url,
                                 data=json_str,
                                 headers={'Content-Type': 'application/json'}
                                 )
        if not response.ok:
            raise Exception(f'Cannot access config server: {response.text}')
    except Exception as e:
        logger.info(f'Problem with set_config: {e!r}')
    else:
        if parse:
            config_entry = from_json(response.content.decode('utf8'))
        else:
            config_entry = response.json()

    return config_entry



import os
import numpy as np
import pytest
import shutil
import tempfile

from astropy.nddata import Cutout2D

from panoptes.utils import images as img_utils
from panoptes.utils import error


def test_crop_data():
    ones = np.ones((201, 201))
    assert ones.sum() == 40401.

    cropped01 = img_utils.crop_data(ones)  # False to exercise coverage.
    assert cropped01.sum() == 40000.

    cropped02 = img_utils.crop_data(ones, box_width=10)
    assert cropped02.sum() == 100.

    cropped03 = img_utils.crop_data(ones, box_width=6, center=(50, 50))
    assert cropped03.sum() == 36.

    # Test the Cutout2D object
    cropped04 = img_utils.crop_data(ones,
                                    box_width=20,
                                    center=(50, 50),
                                    data_only=False)
    assert isinstance(cropped04, Cutout2D)
    assert cropped04.position_original == (50, 50)

    # Box is 20 pixels wide so center is at 10,10
    assert cropped04.position_cutout == (10, 10)


def test_make_pretty_image(solved_fits_file, tiny_fits_file, save_environ):
    # Make a dir and put test image files in it.
    with tempfile.TemporaryDirectory() as tmpdir:
        fits_file = os.path.join(tmpdir, os.path.basename(tiny_fits_file))
        # TODO Add a small CR2 file to our sample image files.

        # Can't operate on a non-existent files.
        with pytest.warns(UserWarning, match="File doesn't exist"):
            assert not img_utils.make_pretty_image(fits_file)

        # Copy the files.
        shutil.copy(solved_fits_file, tmpdir)
        shutil.copy(tiny_fits_file, tmpdir)

        # Can handle the fits file, and creating the images dir for linking
        # the latest image.
        imgdir = os.path.join(tmpdir, 'images')
        assert not os.path.isdir(imgdir)
        os.makedirs(imgdir, exist_ok=True)
        os.environ['PANDIR'] = tmpdir

        link_path = os.path.expandvars('$PANDIR/latest.jpg')
        pretty = img_utils.make_pretty_image(fits_file, link_path=link_path)
        assert pretty
        assert os.path.isfile(pretty)
        assert os.path.isdir(imgdir)
        assert link_path == pretty
        os.remove(link_path)
        os.rmdir(imgdir)

        # Try again, but without link_path.
        pretty = img_utils.make_pretty_image(fits_file, title='some text')
        assert pretty
        assert os.path.isfile(pretty)
        assert not os.path.isdir(imgdir)


@pytest.mark.skipif(
    "TRAVIS" in os.environ and os.environ["TRAVIS"] == "true",
    reason="Skipping this test on Travis CI.")
def test_make_pretty_image_cr2_fail():
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpfile = os.path.join(tmpdir, 'bad.cr2')
        with open(tmpfile, 'w') as f:
            f.write('not an image file')
        with pytest.raises(error.InvalidCommand):
            img_utils.make_pretty_image(tmpfile,
                                        title='some text')
        with pytest.raises(error.InvalidCommand):
            img_utils.make_pretty_image(tmpfile)


@pytest.mark.skipif("TRAVIS" not in os.environ, reason="Skipping this test if not on Travis CI.")
def test_make_pretty_image_cr2(cr2_file):
    link_path = os.path.expandvars('$PANDIR/images/latest.jpg')
    pretty_path = img_utils.make_pretty_image(cr2_file,
                                              title='CR2 Test',
                                              image_type='cr2',
                                              link_path=link_path)

    assert os.path.exists(pretty_path)
    assert pretty_path == link_path



import threading
import weakref
from uuid import uuid4
from contextlib import suppress

from .. import error
from ..serializers import to_json
from ..serializers import from_json
from ..database import AbstractPanDB
from ..database import create_storage_obj


class PanMemoryDB(AbstractPanDB):
    """In-memory store of serialized objects.

    We serialize the objects in order to test the same code path used
    when storing in an external database.
    """

    active_dbs = weakref.WeakValueDictionary()

    @classmethod
    def get_or_create(cls, db_name=None, **kwargs):
        """Returns the named db, creating if needed.

        This method exists because PanDB gets called multiple times for
        the same database name. With mongo or a file store where the storage
        is external from the instance, that is not a problem, but with
        PanMemoryDB the instance is the store, so the instance must be
        shared."""
        db = PanMemoryDB.active_dbs.get(db_name)
        if not db:
            db = PanMemoryDB(db_name=db_name, **kwargs)
            PanMemoryDB.active_dbs[db_name] = db
        return db

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.current = {}
        self.collections = {}
        self.lock = threading.Lock()

    def _make_id(self):
        return str(uuid4())

    def insert_current(self, collection, obj, store_permanently=True):
        self.validate_collection(collection)
        obj_id = self._make_id()
        obj = create_storage_obj(collection, obj, obj_id=obj_id)
        try:
            obj = to_json(obj)
        except Exception as e:
            raise error.InvalidSerialization(f"Problem serializing object for insertion: {e} {obj!r}")

        with self.lock:
            self.current[collection] = obj
            if store_permanently:
                self.collections.setdefault(collection, {})[obj_id] = obj
        return obj_id

    def insert(self, collection, obj):
        self.validate_collection(collection)
        obj_id = self._make_id()
        obj = create_storage_obj(collection, obj, obj_id=obj_id)
        try:
            obj = to_json(obj)
        except Exception as e:
            raise error.InvalidSerialization(f"Problem inserting object into collection: {e}, {obj!r}")

        with self.lock:
            self.collections.setdefault(collection, {})[obj_id] = obj
        return obj_id

    def get_current(self, collection):
        with self.lock:
            obj = self.current.get(collection, None)
        if obj:
            obj = from_json(obj)
        return obj

    def find(self, collection, obj_id):
        with self.lock:
            obj = self.collections.get(collection, {}).get(obj_id)
        if obj:
            obj = from_json(obj)
        return obj

    def clear_current(self, entry_type):
        with suppress(KeyError):
            del self.current[entry_type]

    @classmethod
    def permanently_erase_database(self, db_name):
        # For some reason we're not seeing all the references disappear
        # after tests. Perhaps there is some global variable pointing at
        # the db or one of its referrers, or perhaps a pytest fixture
        # hasn't been removed.
        PanMemoryDB.active_dbs = weakref.WeakValueDictionary()



import sys
from .logger import logger


class PanError(Exception):

    """ Base class for Panoptes errors """

    def __init__(self, msg=None, exit=False):
        self.msg = msg

        if self.msg:
            logger.error(str(self))

        if exit:
            self.exit_program(self.msg)

    def exit_program(self, msg=None):
        """ Kills running program """
        if not msg:
            self.msg = 'No reason specified'

        # Show on current display.
        logger.add(sys.stderr, format='<red><b>{message}</></>')
        logger.critical(f"TERMINATING: {self.msg}")
        sys.exit(1)

    def __str__(self):
        error_str = str(self.__class__.__name__)
        if self.msg:
            error_str += ': {}'.format(self.msg)

        return error_str


class InvalidSystemCommand(PanError):

    """ Error for a system level command malfunction """

    def __init__(self, msg='Problem running system command', **kwargs):
        super().__init__(msg, **kwargs)


class InvalidDeserialization(PanError):

    """ Error for serialization errors """

    def __init__(self, msg='Problem deserializing', **kwargs):
        super().__init__(msg, **kwargs)


class InvalidSerialization(PanError):

    """ Error for serialization errors """

    def __init__(self, msg='Problem Serializing', **kwargs):
        super().__init__(msg, **kwargs)


class Timeout(PanError):

    """ Error called when an event times out """

    def __init__(self, msg='Timeout waiting for event', **kwargs):
        super().__init__(msg, **kwargs)


class NoObservation(PanError):

    """ Generic no Observation """

    def __init__(self, msg='No valid observations found.', **kwargs):
        super().__init__(msg, **kwargs)


class NotFound(PanError):

    """ Generic not found class """
    pass


class InvalidCollection(NotFound):
    """PanError raised if a collection name is invalid."""
    pass


class InvalidConfig(PanError):

    """ PanError raised if config file is invalid """
    pass


class InvalidCommand(PanError):

    """ PanError raised if a system command does not run """
    pass


class InvalidMountCommand(PanError):

    """ PanError raised if attempting to send command that doesn't exist """
    pass


class InvalidObservation(NotFound):
    """PanError raised if a field is invalid."""
    pass


class BadConnection(PanError):

    """ PanError raised when a connection is bad """
    pass


class BadSerialConnection(PanError):

    """ PanError raised when serial command is bad """
    pass


class ArduinoDataError(PanError):
    """PanError raised when there is something very wrong with Arduino information."""
    pass


class MountNotFound(NotFound):

    """ Mount cannot be import """

    def __init__(self, msg='Mount Not Found', **kwargs):
        super().__init__(msg, **kwargs)


class CameraNotFound(NotFound):

    """ Camera cannot be imported """
    pass


class DomeNotFound(NotFound):
    """Dome device not found."""
    pass


class SolveError(NotFound):

    """ Camera cannot be imported """
    pass


class TheSkyXError(PanError):
    """ Errors from TheSkyX """
    pass


class TheSkyXKeyError(TheSkyXError):
    """ Errors from TheSkyX because bad key passed """
    pass


class TheSkyXTimeout(TheSkyXError):
    """ Errors from TheSkyX because bad key passed """
    pass


class GoogleCloudError(PanError):
    """ Errors related to google cloud """
    pass


class NotSupported(PanError, NotImplementedError):
    """ Errors from trying to use hardware features not supported by a particular model """
    pass


class IllegalValue(PanError, ValueError):
    """ Errors from trying to hardware parameters to values not supported by a particular model """
    pass



import io
import pytest
import serial
from serial import serialutil

from panoptes.utils import error
from panoptes.utils import rs232

from panoptes.utils.tests.serial_handlers import NoOpSerial
from panoptes.utils.tests.serial_handlers import protocol_buffers
from panoptes.utils.tests.serial_handlers import protocol_hooked


def test_port_discovery():
    ports = rs232.get_serial_port_info()
    assert isinstance(ports, list)


def test_missing_port():
    with pytest.raises(ValueError):
        rs232.SerialData()


def test_non_existent_device():
    """Doesn't complain if it can't find the device."""
    port = '/dev/tty12345698765'
    ser = rs232.SerialData(port=port)
    assert not ser.is_connected
    assert port == ser.name
    # Can't connect to that device.
    with pytest.raises(error.BadSerialConnection):
        ser.connect()
    assert not ser.is_connected


def test_detect_uninstalled_scheme():
    """If our handlers aren't installed, will detect unknown scheme."""
    # See https://pythonhosted.org/pyserial/url_handlers.html#urls for info on the
    # standard schemes that are supported by PySerial.
    with pytest.raises(ValueError):
        # The no_op scheme references one of our test handlers, but it shouldn't be
        # accessible unless we've added our package to the list to be searched.
        rs232.SerialData(port='no_op://')


@pytest.fixture(scope='function')
def handler():
    # Install our package that contain the test handlers.
    serial.protocol_handler_packages.append('panoptes.utils.tests.serial_handlers')
    yield True
    # Remove that package.
    serial.protocol_handler_packages.remove('panoptes.utils.tests.serial_handlers')


def test_detect_bogus_scheme(handler):
    """When our handlers are installed, will still detect unknown scheme."""
    with pytest.raises(ValueError) as excinfo:
        # The scheme (the part before the ://) must be a Python module name, so use
        # a string that can't be a module name.
        rs232.SerialData(port='# bogus #://')
    assert '# bogus #' in repr(excinfo.value)


def test_custom_logger(handler, fake_logger):
    s0 = rs232.SerialData(port='no_op://', logger=fake_logger)
    s0.logger.debug('Testing logger')


def test_basic_no_op(handler):
    # Confirm we can create the SerialData object.
    ser = rs232.SerialData(port='no_op://', name='a name', open_delay=0)
    assert ser.name == 'a name'

    # Peek inside, it should have a NoOpSerial instance as member ser.
    assert ser.ser
    assert isinstance(ser.ser, NoOpSerial)

    # Open is automatically called by SerialData.
    assert ser.is_connected

    # connect() is idempotent.
    ser.connect()
    assert ser.is_connected

    # Several passes of reading, writing, disconnecting and connecting.
    for _ in range(3):
        # no_op handler doesn't do any reading, analogous to /dev/null, which
        # never produces any output.
        assert '' == ser.read(retry_delay=0.01, retry_limit=2)
        assert b'' == ser.read_bytes(size=1)
        assert 0 == ser.write('abcdef')
        ser.reset_input_buffer()

        # Disconnect from the serial port.
        assert ser.is_connected
        ser.disconnect()
        assert not ser.is_connected

        # Should no longer be able to read or write.
        with pytest.raises(AssertionError):
            ser.read(retry_delay=0.01, retry_limit=1)
        with pytest.raises(AssertionError):
            ser.read_bytes(size=1)
        with pytest.raises(AssertionError):
            ser.write('a')
        ser.reset_input_buffer()

        # And we should be able to reconnect.
        assert not ser.is_connected
        ser.connect()
        assert ser.is_connected


def test_basic_io(handler):
    protocol_buffers.ResetBuffers(b'abc\r\ndef\n')
    ser = rs232.SerialData(port='buffers://', open_delay=0.01, retry_delay=0.01,
                           retry_limit=2)

    # Peek inside, it should have a BuffersSerial instance as member ser.
    assert isinstance(ser.ser, protocol_buffers.BuffersSerial)

    # Can read two lines. Read the first as a sensor reading:
    (ts, line) = ser.get_reading()
    assert 'abc\r\n' == line

    # Read the second line from the read buffer.
    assert 'def\n' == ser.read(retry_delay=0.1, retry_limit=10)

    # Another read will fail, having exhausted the contents of the read buffer.
    assert '' == ser.read()

    # Can write to the "device", the handler will accumulate the results.
    assert 5 == ser.write('def\r\n')
    assert 6 == ser.write('done\r\n')

    assert b'def\r\ndone\r\n' == protocol_buffers.GetWBufferValue()

    # If we add more to the read buffer, we can read again.
    protocol_buffers.SetRBufferValue(b'line1\r\nline2\r\ndangle')
    assert 'line1\r\n' == ser.read(retry_delay=10, retry_limit=20)
    assert 'line2\r\n' == ser.read(retry_delay=10, retry_limit=20)
    assert 'dangle' == ser.read(retry_delay=10, retry_limit=20)

    ser.disconnect()
    assert not ser.is_connected


class HookedSerialHandler(NoOpSerial):
    """Sources a line of text repeatedly, and sinks an infinite amount of input."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.r_buffer = io.BytesIO(
            b"{'a': 12, 'b': [1, 2, 3, 4], 'c': {'d': 'message'}}\r\n")

    @property
    def in_waiting(self):
        """The number of input bytes available to read immediately."""
        if not self.is_open:
            raise serialutil.portNotOpenError
        total = len(self.r_buffer.getbuffer())
        avail = total - self.r_buffer.tell()
        # If at end of the stream, reset the stream.
        if avail <= 0:
            self.r_buffer.seek(0)
            avail = total
        return avail

    def open(self):
        """Open port.

        Raises:
            SerialException if the port cannot be opened.
        """
        self.is_open = True

    def close(self):
        """Close port immediately."""
        self.is_open = False

    def read(self, size=1):
        """Read until the end of self.r_buffer, then seek to beginning of self.r_buffer."""
        if not self.is_open:
            raise serialutil.portNotOpenError
        # If at end of the stream, reset the stream.
        return self.r_buffer.read(min(size, self.in_waiting))

    def write(self, data):
        """Write data to bitbucket."""
        if not self.is_open:
            raise serialutil.portNotOpenError
        return len(data)


def test_hooked_io(handler):
    protocol_hooked.Serial = HookedSerialHandler
    ser = rs232.SerialData(port='hooked://', open_delay=0)

    # Peek inside, it should have a PySerial instance as member ser.
    assert ser.ser
    assert ser.ser.__class__.__name__ == 'HookedSerialHandler'
    print(str(ser.ser))

    # Open is automatically called by SerialData.
    assert ser.is_connected

    # Can read many identical lines from ser.
    first_line = None
    for n in range(20):
        line = ser.read(retry_delay=10, retry_limit=20)
        if first_line:
            assert line == first_line
        else:
            first_line = line
            assert 'message' in line
        reading = ser.get_reading()
        assert reading[1] == line

    # Can write to the "device" many times.
    line = 'abcdefghijklmnop' * 30
    line = line + '\r\n'
    for n in range(20):
        assert len(line) == ser.write(line)

    ser.disconnect()
    assert not ser.is_connected



import pytest

from mocket import Mocket

from panoptes.utils import error
from panoptes.utils.theskyx import TheSkyX


@pytest.fixture(scope="function")
def skyx(data_dir, request):
    """Create TheSkyX class but don't connect.t

    If running with a real connection TheSkyX then the Mokcet will
    be disabled here.
    """

    # Use `--with-hardware thesky` on cli to run without mock
    Mocket.enable('theskyx', data_dir)
    if 'theskyx' in request.config.getoption('--with-hardware'):
        Mocket.disable()

    theskyx = TheSkyX(connect=False)

    yield theskyx


def test_default_connect(data_dir, request):
    """Test connection to TheSkyX

    If not running with a real connection then use Mocket
    """
    # Use `--with-hardware thesky` on cli to run without mock
    if 'theskyx' not in request.config.getoption('--with-hardware'):
        Mocket.enable('theskyx', data_dir)

    skyx = TheSkyX()
    assert skyx.is_connected is True


def test_no_connect_write(skyx):
    with pytest.raises(error.BadConnection):
        skyx.write('/* Java Script */')


def test_no_connect_read(skyx):
    with pytest.raises(error.BadConnection):
        skyx.read()


def test_write_bad_key(skyx):
    skyx.connect()
    skyx.write('FOOBAR')
    with pytest.raises(error.TheSkyXKeyError):
        skyx.read()


def test_write_no_command(skyx):
    skyx.connect()
    skyx.write('/* Java Script */')
    assert skyx.read() == 'undefined'


def test_get_build(skyx):
    js = '''
/* Java Script */
var Out;
Out=Application.version
'''
    skyx.connect()
    skyx.write(js)
    assert skyx.read().startswith('10.5')


def test_error(skyx):
    skyx.connect()
    skyx.write('''
/* Java Script */
sky6RASCOMTele.FindHome()
''')
    with pytest.raises(error.TheSkyXError):
        skyx.read()



import re
import zmq

from .logger import logger
from .time import current_time
from .serializers import from_json
from .serializers import to_json


class PanMessaging(object):
    """Provides messaging services within a PANOPTES robotic telescope.

    Supports broadcasting messages from publishers (e.g. a POCS or
    ArduinoIO class instance) to subscribers (also typically class
    instances). The publishers and subscribers may be in the same
    process, or in separate processes. The messages all go through
    a message forwarder; this is a process which listens for messages
    from all publishers on one TCP port and forwards each message to
    all subscribers that are connected to a second TCP port.

    Do not create PanMessaging instances directly. Publishers should
    call PanMessaging.create_publisher to create an instance of
    PanMessaging, on which they can then call send_message.
    Subscribers should call PanMessaging.create_subscriber to create
    an instance of PanMessaging, on which they can then call
    receive_message.

    Messages are sent to topics, a name that can be used to allow
    a high-level partitioning of messages. A topic name may not
    include whitespace. Among the currently used topic names are:

      * PANCHAT (sent from POCS.say)
      * PAWS-CMD (sent from PAWS websockets.py)
      * POCS (sent by class POCS)
      * POCS-CMD (sent by class POCS)
      * STATUS (sent by class POCS)
      * weather (from peas/sensors.py)
      * environment (from peas/sensors.py)
      * telemetry:commands (in ArduinoIO... new)
      * camera:commands (in ArduinoIO... new)

    And some other topics are used in tests:

      * Test-Topic (test_messaging.py)
      * RUNNING (test_pocs.py)
      * POCS-CMD (test_pocs.py)

    The method receive_message will return messages from all topics;
    the caller must check the returned topic name to determine if
    the message value is of interest.

    Note: PAWS doesn't use PanMessaging, which will likely result in
    problems as we evolve PanMessaging and the set of topics.
    TODO: Figure out how to share PanMessaging with PAWS.

    The value of a message being sent may be a string (in which case it
    is wrapped in a dict(message=<value>, timestamp=<now>) or a dict,
    in which case it will be "scrubbed", i.e. the dict entries will be
    modified as necessary to so that the dict can be serialized using
    json.dumps.

    TODO Pick an encoding of strings (e.g. UTF-8) so that non-ASCII
    strings may be sent and received without corruption of the data
    or exceptions being thrown.

    ZeroMQ is used to provide the underlying pub-sub support. ZeroMQ
    supports only a very basic message format: an array of bytes.
    PanMessaging converts the provided message topic and value into
    a byte array of this format:
        <topic-name><space><serialized-value>
    """
    logger = logger

    # Topic names must consist of the characters.
    topic_name_re = re.compile('[a-zA-Z][-a-zA-Z0-9_.:]*')

    def __init__(self, **kwargs):
        """Do not call this directly."""
        # Create a new context
        self.context = zmq.Context()
        self.socket = None

    @classmethod
    def create_forwarder(cls, sub_port, pub_port, ready_fn=None, done_fn=None):
        subscriber, publisher = PanMessaging.create_forwarder_sockets(sub_port, pub_port)
        PanMessaging.run_forwarder(subscriber, publisher, ready_fn=ready_fn, done_fn=done_fn)

    @classmethod
    def create_forwarder_sockets(cls, sub_port, pub_port):
        cls.logger.info('Creating forwarder sockets for {} -> {}', sub_port, pub_port)
        subscriber = PanMessaging.create_subscriber(sub_port, bind=True, connect=False)
        publisher = PanMessaging.create_publisher(pub_port, bind=True, connect=False)
        return subscriber, publisher

    @classmethod
    def run_forwarder(cls, subscriber, publisher, ready_fn=None, done_fn=None):
        publisher.logger.info('run_forwarder')
        try:
            if ready_fn:
                ready_fn()
            publisher.logger.info('run_forwarder calling zmq.device')
            zmq.device(zmq.FORWARDER, subscriber.socket, publisher.socket)
        except KeyboardInterrupt:
            pass
        except Exception as e:
            publisher.logger.warning(e)
            publisher.logger.warning("bringing down zmq device")
        finally:
            publisher.logger.info('run_forwarder closing publisher and subscriber')
            publisher.close()
            subscriber.close()
            if done_fn:
                done_fn()

    @classmethod
    def create_publisher(cls, port, bind=False, host='localhost', connect=True):
        """ Create a publisher

        Args:
            port (int): The port (on localhost) to bind to.

        Returns:
            A ZMQ PUB socket
        """
        obj = cls()

        obj.logger.debug('Creating zmq message publisher.')

        socket = obj.context.socket(zmq.PUB)

        if bind:
            obj.logger.debug(f'Binding publisher to port {port}')
            socket.bind(f'tcp://*:{port}')
        elif connect:
            obj.logger.debug(f'Binding publisher to tcp://{host}:{port}')
            socket.connect(f'tcp://{host}:{port}')

        obj.socket = socket

        return obj

    @classmethod
    def create_subscriber(cls, port, topic='', host='localhost', bind=False, connect=True):
        """ Create a listener

        Args:
            port (int):         The port (on localhost) to bind to.
            topic (str):      Which topic or topic prefix to subscribe to.

        """
        obj = cls()
        obj.logger.debug("Creating subscriber. Port: {} \tTopic: {}".format(port, topic))

        socket = obj.context.socket(zmq.SUB)

        if bind:
            try:
                socket.bind(f'tcp://*:{port}')
            except zmq.error.ZMQError:
                obj.logger.debug('Problem binding port {}'.format(port))
        elif connect:
            obj.logger.debug(f'Connecting subscriber to tcp://{host}:{port}')
            socket.connect(f'tcp://{host}:{port}')

        socket.setsockopt_string(zmq.SUBSCRIBE, topic)

        obj.socket = socket

        return obj

    def send_message(self, topic, message):
        """ Responsible for actually sending message across a topic

        Args:
            topic(str):   Name of topic to send on. The name must
                match topic_name_re.
            message:   Message to be sent (a string or a dict).
        """
        if not isinstance(topic, str):
            raise ValueError('Topic name must be a string')
        elif not self.topic_name_re.fullmatch(topic):
            raise ValueError('Topic name ("{}") is not valid'.format(topic))

        if topic == 'PANCHAT':
            self.logger.info(f"{topic} {message}")

        if isinstance(message, str):
            message = to_json({
                "message": message,
                "timestamp": current_time(pretty=True),
            })
        elif isinstance(message, dict):
            message = to_json(message)
        else:
            raise ValueError('Message value must be a string or dict')

        # Build the full message with topic
        full_message = f'{topic} {message}'

        # Send the message
        # self.socket.send_string(full_message, flags=zmq.NOBLOCK)
        self.socket.send_string(full_message)

    def receive_message(self, blocking=True, flags=0, timeout_ms=0):
        """Receive a message

        Receives a message for the current subscriber. Blocks by default, pass
        `flags=zmq.NOBLOCK` for non-blocking.

        Args:
            blocking (bool, optional): If True, blocks until message
                received or timeout__ms elapsed (if timeout_ms > 0).
            flag (int, optional): Any valid recv flag, e.g. zmq.NOBLOCK
            timeout_ms (int, optional): Time in milliseconds to wait for
                a message to arrive. Only applies if blocking is True.

        Returns:
            tuple(str, dict): Tuple containing the topic and a dict
        """
        topic = None
        msg_obj = None
        if not blocking:
            flags = flags | zmq.NOBLOCK
        elif timeout_ms > 0:
            # Wait until a message is available or the timeout expires.
            # TODO(jamessynge): Remove flags=..., confirm that works with
            # the default flags value of zmq.POLLIN.
            self.socket.poll(timeout=timeout_ms, flags=(zmq.POLLIN | zmq.POLLOUT))
            # Don't block at this point, because we will have waited as long
            # as necessary.
            flags = flags | zmq.NOBLOCK

        try:
            # Ugh. So ugly with the strings.
            message = self.socket.recv_string(flags=flags)
        except Exception as e:
            self.logger.warning(f'error in receive_message: {e!r}')
        else:
            topic, msg = message.split(' ', maxsplit=1)
            try:
                msg_obj = from_json(msg)
            except Exception as e:
                self.logger.warning(f'Error parsing message: {msg} {e!r}')

        return topic, msg_obj

    def close(self):
        """Close the socket """
        self.socket.close()
        self.context.term()



from copy import copy

from warnings import warn

from matplotlib import rc
from matplotlib import animation
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.figure import Figure
from matplotlib import cm
from mpl_toolkits.axes_grid1 import make_axes_locatable

import numpy as np

from astropy.visualization import LogStretch, ImageNormalize, LinearStretch, MinMaxInterval

rc('animation', html='html5')


def get_palette(cmap='inferno'):
    """Get a palette for drawing.

    Returns a copy of the colormap palette with bad pixels marked.

    Args:
        cmap (str, optional): Colormap to use, default 'inferno'.

    Returns:
        `matplotlib.cm`: The colormap.
    """
    palette = copy(getattr(cm, cmap))

    # Mark bad pixels (e.g. saturated)
    # when using vmin or vmax and a normalizer.
    palette.set_over('w', 1.0)
    palette.set_under('k', 1.0)
    palette.set_bad('g', 1.0)

    return palette


def add_colorbar(axes_image, size='5%', pad=0.05, orientation='vertical'):
    """Add a colorbar to the image.

    This is a simple convenience function to add a colorbar to a plot generated by
    `matplotlib.pyplot.imshow`.

    .. plot::

        >>> from matplotlib import pyplot as plt
        >>> import numpy as np
        >>> from panoptes.utils.images.plot import add_colorbar
        >>>
        >>> x = np.arange(0.0, 100.0)
        >>> y = np.arange(0.0, 100.0)
        >>> X, Y = np.meshgrid(x, y)
        >>>
        >>> func = lambda x, y: x**2 + y**2
        >>>
        >>> z = func(X, Y)
        >>>
        >>> fig, ax = plt.subplots()
        >>> im1 = ax.imshow(z, origin='lower')
        >>> add_colorbar(im1)
        >>> fig.show()


    Args:
        axes_image (`matplotlib.image.AxesImage`): A matplotlib AxesImage.
    """
    divider = make_axes_locatable(axes_image.axes)
    cax = divider.append_axes('right', size=size, pad=pad)
    axes_image.figure.colorbar(axes_image, cax=cax, orientation=orientation)


def add_pixel_grid(ax1, grid_height, grid_width, show_axis_labels=True, show_superpixel=False,
                   major_alpha=0.5, minor_alpha=0.25):

    # major ticks every 2, minor ticks every 1
    if show_superpixel:
        x_major_ticks = np.arange(-0.5, grid_width, 2)
        y_major_ticks = np.arange(-0.5, grid_height, 2)

        ax1.set_xticks(x_major_ticks)
        ax1.set_yticks(y_major_ticks)

        ax1.grid(which='major', color='r', linestyle='--', lw=3, alpha=major_alpha)
    else:
        ax1.set_xticks([])
        ax1.set_yticks([])

    x_minor_ticks = np.arange(-0.5, grid_width, 1)
    y_minor_ticks = np.arange(-0.5, grid_height, 1)

    ax1.set_xticks(x_minor_ticks, minor=True)
    ax1.set_yticks(y_minor_ticks, minor=True)

    ax1.grid(which='minor', color='r', lw='2', linestyle='--', alpha=minor_alpha)

    if show_axis_labels is False:
        ax1.set_xticklabels([])
        ax1.set_yticklabels([])


def animate_stamp(d0):

    fig = Figure()
    FigureCanvas(fig)

    ax = fig.add_subplot(111)
    ax.set_xticks([])
    ax.set_yticks([])

    line = ax.imshow(d0[0])
    ax.set_title(f'Frame 0')

    def animate(i):
        line.set_data(d0[i])  # update the data
        ax.set_title(f'Frame {i:03d}')
        return line,

    # Init only required for blitting to give a clean slate.
    def init():
        line.set_data(d0[0])
        return line,

    ani = animation.FuncAnimation(fig, animate, np.arange(0, len(d0)), init_func=init,
                                  interval=500, blit=True)

    return ani


def show_stamps(pscs,
                frame_idx=None,
                stamp_size=11,
                aperture_position=None,
                show_residual=False,
                stretch=None,
                save_name=None,
                show_max=False,
                show_pixel_grid=False,
                **kwargs):

    if aperture_position is None:
        midpoint = (stamp_size - 1) / 2
        aperture_position = (midpoint, midpoint)

    ncols = len(pscs)

    if show_residual:
        ncols += 1

    nrows = 1

    fig = Figure()
    FigureCanvas(fig)
    fig.set_figheight(4)
    fig.set_figwidth(8)

    if frame_idx is not None:
        s0 = pscs[0][frame_idx]
        s1 = pscs[1][frame_idx]
    else:
        s0 = pscs[0]
        s1 = pscs[1]

    if stretch == 'log':
        stretch = LogStretch()
    else:
        stretch = LinearStretch()

    norm = ImageNormalize(s0, interval=MinMaxInterval(), stretch=stretch)

    ax1 = fig.add_subplot(nrows, ncols, 1)

    im = ax1.imshow(s0, cmap=get_palette(), norm=norm)

    # create an axes on the right side of ax. The width of cax will be 5%
    # of ax and the padding between cax and ax will be fixed at 0.05 inch.
    # https://stackoverflow.com/questions/18195758/set-matplotlib-colorbar-size-to-match-graph
    divider = make_axes_locatable(ax1)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    fig.colorbar(im, cax=cax)
    ax1.set_title('Target')

    # Comparison
    ax2 = fig.add_subplot(nrows, ncols, 2)
    im = ax2.imshow(s1, cmap=get_palette(), norm=norm)

    divider = make_axes_locatable(ax2)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    fig.colorbar(im, cax=cax)
    ax2.set_title('Comparison')

    if show_pixel_grid:
        add_pixel_grid(ax1, stamp_size, stamp_size, show_superpixel=False)
        add_pixel_grid(ax2, stamp_size, stamp_size, show_superpixel=False)

    if show_residual:
        ax3 = fig.add_subplot(nrows, ncols, 3)

        # Residual
        residual = s0 - s1
        im = ax3.imshow(residual, cmap=get_palette(), norm=ImageNormalize(
            residual, interval=MinMaxInterval(), stretch=LinearStretch()))

        divider = make_axes_locatable(ax3)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        fig.colorbar(im, cax=cax)
        ax3.set_title('Noise Residual')
        ax3.set_title('Residual RMS: {:.01%}'.format(residual.std()))
        ax3.set_yticklabels([])
        ax3.set_xticklabels([])

        if show_pixel_grid:
            add_pixel_grid(ax1, stamp_size, stamp_size, show_superpixel=False)

    # Turn off tick labels
    ax1.set_yticklabels([])
    ax1.set_xticklabels([])
    ax2.set_yticklabels([])
    ax2.set_xticklabels([])

    if save_name:
        try:
            fig.savefig(save_name)
        except Exception as e:
            warn("Can't save figure: {}".format(e))

    return fig



import os
import shutil
import subprocess
from warnings import warn

from astropy import units as u
from astropy.io import fits
from astropy.wcs import WCS

from .. import error
from ..logger import logger


def solve_field(fname, timeout=15, solve_opts=None, *args, **kwargs):
    """ Plate solves an image.

    Note: This is a low-level wrapper around the underlying `solve-field`
        program. See `get_solve_field` for more typical usage and examples.


    Args:
        fname(str, required):       Filename to solve in .fits extension.
        timeout(int, optional):     Timeout for the solve-field command,
                                    defaults to 60 seconds.
        solve_opts(list, optional): List of options for solve-field.
    """
    solve_field_script = shutil.which('panoptes-solve-field')

    if solve_field_script is None:  # pragma: no cover
        raise error.InvalidSystemCommand(
            "Can't find panoptes-solve-field: {}".format(solve_field_script))

    # Add the options for solving the field
    if solve_opts is not None:
        options = solve_opts
    else:
        # Default options
        options = [
            '--guess-scale',
            '--cpulimit', str(timeout),
            '--no-verify',
            '--crpix-center',
            '--temp-axy',
            '--index-xyls', 'none',
            '--solved', 'none',
            '--match', 'none',
            '--rdls', 'none',
            '--corr', 'none',
            '--downsample', '4',
            '--no-plots',
        ]

        if 'ra' in kwargs:
            options.append('--ra')
            options.append(str(kwargs.get('ra')))
        if 'dec' in kwargs:
            options.append('--dec')
            options.append(str(kwargs.get('dec')))
        if 'radius' in kwargs:
            options.append('--radius')
            options.append(str(kwargs.get('radius')))

    # Gather all the kwargs that start with `--` and are not already present.
    logger.debug(f'Adding args: {args!r}')
    options.extend([opt
                    for opt
                    in args
                    if opt.startswith('--') and opt not in options])
    logger.debug(f'Adding kwargs: {kwargs!r}')
    options.extend([f'{opt}={val}'
                    for opt, val
                    in kwargs.items()
                    if opt.startswith('--') and opt not in options])

    cmd = [solve_field_script] + options + [fname]

    logger.debug(f'Solving with: {cmd}')
    try:
        proc = subprocess.Popen(cmd,
                                universal_newlines=True,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
    except Exception as e:
        raise error.PanError("Problem plate-solving in solve_field: {}".format(e))

    return proc


def get_solve_field(fname, replace=True, overwrite=True, *args, **kwargs):
    """Convenience function to wait for `solve_field` to finish.

    This function merely passes the `fname` of the image to be solved along to `solve_field`,
    which returns a subprocess.Popen object. This function then waits for that command
    to complete, populates a dictonary with the EXIF informaiton and returns. This is often
    more useful than the raw `solve_field` function

    Example:

    >>> from panoptes.utils.images import fits as fits_utils

    >>> # Get our fits filename.
    >>> fits_fn = getfixture('unsolved_fits_file')

    >>> # Perform the solve.
    >>> solve_info = fits_utils.get_solve_field(fits_fn)

    >>> # Show solved filename.
    >>> solve_info['solved_fits_file']
    '.../unsolved.fits'

    >>> # Pass a suggested location.
    >>> ra = 15.23
    >>> dec = 90
    >>> radius = 5 # deg
    >>> solve_info = fits_utils.solve_field(fits_fn, ra=ra, dec=dec, radius=radius)

    >>> # Pass args and kwargs to `solve-field` program.
    >>> solve_args = ['--use-wget']
    >>> solve_kwargs = {'--pnm': './awesome.bmp'}
    >>> solve_info = fits_utils.get_solve_field(fits_fn, *solve_args, **solve_kwargs, skip_solved=False)

    Args:
        fname ({str}): Name of FITS file to be solved.
        replace (bool, optional): Saves the WCS back to the original file,
            otherwise output base filename with `.new` extension. Default True.
        overwrite (bool, optional): Clobber file, default True. Required if `replace=True`.
        **args ({dict}): Options to pass to `solve_field` should start with `--`.
        **kwargs ({dict}): Options to pass to `solve_field` should start with `--`.

    Returns:
        dict: Keyword information from the solved field.
    """
    skip_solved = kwargs.get('skip_solved', True)

    out_dict = {}
    output = None
    errs = None

    header = getheader(fname)
    wcs = WCS(header)

    # Check for solved file
    if skip_solved and wcs.is_celestial:
        logger.info(f"Skipping solved file (use skip_solved=False to solve again): {fname}")

        out_dict.update(header)
        out_dict['solved_fits_file'] = fname
        return out_dict

    # Set a default radius of 15
    if overwrite:
        args = list(args)
        args.append('--overwrite')

    # Use unpacked version of file.
    was_compressed = False
    if fname.endswith('.fz'):
        logger.debug(f'Uncompressing {fname}')
        fname = funpack(fname)
        logger.debug(f'Using {fname} for solving')
        was_compressed = True

    proc = solve_field(fname, *args, **kwargs)
    try:
        output, errs = proc.communicate(timeout=kwargs.get('timeout', 30))
    except subprocess.TimeoutExpired:
        proc.kill()
        output, errs = proc.communicate()
        raise error.Timeout(f'Timeout while solving: {output!r} {errs!r}')
    else:
        if proc.returncode != 0:
            logger.debug(f'Returncode: {proc.returncode}')
        for log in [output, errs]:
            if log and log > '':
                logger.debug(f'Output on {fname}: {log}')

        if proc.returncode == 3:
            raise error.SolveError(f'solve-field not found: {output}')

    new_fname = fname.replace('.fits', '.new')
    if replace:
        logger.debug(f'Overwriting original {fname}')
        os.replace(new_fname, fname)
    else:
        fname = new_fname

    try:
        header = getheader(fname)
        header.remove('COMMENT', ignore_missing=True, remove_all=True)
        header.remove('HISTORY', ignore_missing=True, remove_all=True)
        out_dict.update(header)
    except OSError:
        logger.warning(f"Can't read fits header for: {fname}")

    # Check it was solved.
    if WCS(header).is_celestial is False:
        raise error.SolveError('File not properly solved, no WCS header present.')

    # Remove WCS file.
    os.remove(fname.replace('.fits', '.wcs'))

    if was_compressed and replace:
        logger.debug(f'Compressing plate-solved {fname}')
        fname = fpack(fname)

    out_dict['solved_fits_file'] = fname

    return out_dict


def get_wcsinfo(fits_fname, **kwargs):
    """Returns the WCS information for a FITS file.

    Uses the `wcsinfo` astrometry.net utility script to get the WCS information
    from a plate-solved file.

    Args:
        fits_fname ({str}): Name of a FITS file that contains a WCS.
        **kwargs: Args that can be passed to wcsinfo.

    Returns:
        dict: Output as returned from `wcsinfo`.

    Raises:
        error.InvalidCommand: Raised if `wcsinfo` is not found (part of astrometry.net)
    """
    assert os.path.exists(fits_fname), warn(f"No file exists at: {fits_fname}")

    wcsinfo = shutil.which('wcsinfo')
    if wcsinfo is None:
        raise error.InvalidCommand('wcsinfo not found')

    run_cmd = [wcsinfo, fits_fname]

    if fits_fname.endswith('.fz'):
        run_cmd.append('-e')
        run_cmd.append('1')

    proc = subprocess.Popen(run_cmd, stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT, universal_newlines=True)
    try:
        output, errs = proc.communicate(timeout=5)
    except subprocess.TimeoutExpired:  # pragma: no cover
        proc.kill()
        output, errs = proc.communicate()

    unit_lookup = {
        'crpix0': u.pixel,
        'crpix1': u.pixel,
        'crval0': u.degree,
        'crval1': u.degree,
        'cd11': (u.deg / u.pixel),
        'cd12': (u.deg / u.pixel),
        'cd21': (u.deg / u.pixel),
        'cd22': (u.deg / u.pixel),
        'imagew': u.pixel,
        'imageh': u.pixel,
        'pixscale': (u.arcsec / u.pixel),
        'orientation': u.degree,
        'ra_center': u.degree,
        'dec_center': u.degree,
        'orientation_center': u.degree,
        'ra_center_h': u.hourangle,
        'ra_center_m': u.minute,
        'ra_center_s': u.second,
        'dec_center_d': u.degree,
        'dec_center_m': u.minute,
        'dec_center_s': u.second,
        'fieldarea': (u.degree * u.degree),
        'fieldw': u.degree,
        'fieldh': u.degree,
        'decmin': u.degree,
        'decmax': u.degree,
        'ramin': u.degree,
        'ramax': u.degree,
        'ra_min_merc': u.degree,
        'ra_max_merc': u.degree,
        'dec_min_merc': u.degree,
        'dec_max_merc': u.degree,
        'merc_diff': u.degree,
    }

    wcs_info = {}
    for line in output.split('\n'):
        try:
            k, v = line.split(' ')
            try:
                v = float(v)
            except Exception:
                pass

            wcs_info[k] = float(v) * unit_lookup.get(k, 1)
        except ValueError:
            pass
            # print("Error on line: {}".format(line))

    wcs_info['wcs_file'] = fits_fname

    return wcs_info


def fpack(fits_fname, unpack=False, overwrite=True):
    """Compress/Decompress a FITS file

    Uses `fpack` (or `funpack` if `unpack=True`) to compress a FITS file

    Args:
        fits_fname ({str}): Name of a FITS file that contains a WCS.
        unpack ({bool}, optional): file should decompressed instead of compressed, default False.

    Returns:
        str: Filename of compressed/decompressed file.
    """
    assert os.path.exists(fits_fname), warn(
        "No file exists at: {}".format(fits_fname))

    if unpack:
        fpack = shutil.which('funpack')
        run_cmd = [fpack, '-D', fits_fname]
        out_file = fits_fname.replace('.fz', '')
    else:
        fpack = shutil.which('fpack')
        run_cmd = [fpack, '-D', '-Y', fits_fname]
        out_file = fits_fname.replace('.fits', '.fits.fz')

    if os.path.exists(out_file):
        if overwrite is False:
            raise FileExistsError(f'Destination file already exists at location and overwrite=False')
        else:
            os.remove(out_file)

    try:
        assert fpack is not None
    except AssertionError:
        warn("fpack not found (try installing cfitsio). File has not been changed")
        return fits_fname

    logger.debug("fpack command: {}".format(run_cmd))

    proc = subprocess.Popen(run_cmd, stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT, universal_newlines=True)
    try:
        output, errs = proc.communicate(timeout=5)
    except subprocess.TimeoutExpired:
        proc.kill()
        output, errs = proc.communicate()

    return out_file


def funpack(*args, **kwargs):
    """Unpack a FITS file.

    Note:
        This is a thin-wrapper around the ~fpack function
        with the `unpack=True` option specified. See ~fpack
        documentation for details.

    Args:
        *args: Arguments passed to ~fpack.
        **kwargs: Keyword arguments passed to ~fpack.

    Returns:
        str: Path to uncompressed FITS file.
    """
    return fpack(*args, unpack=True, **kwargs)


def write_fits(data, header, filename, exposure_event=None, **kwargs):
    """Write FITS file to requested location.

    >>> from panoptes.utils.images import fits as fits_utils
    >>> data = np.random.normal(size=100)
    >>> header = { 'FILE': 'delete_me', 'TEST': True }
    >>> filename = str(getfixture('tmpdir').join('temp.fits'))
    >>> fits_utils.write_fits(data, header, filename)
    >>> assert os.path.exists(filename)

    >>> fits_utils.getval(filename, 'FILE')
    'delete_me'
    >>> data2 = fits_utils.getdata(filename)
    >>> assert np.array_equal(data, data2)

    Args:
        data (array_like): The data to be written.
        header (dict): Dictionary of items to be saved in header.
        filename (str): Path to filename for output.
        exposure_event (None|`threading.Event`, optional): A `threading.Event` that
            can be triggered when the image is written.
        kwargs (dict): Options that are passed to the `astropy.io.fits.PrimaryHDU.writeto`
            method.
    """
    if not isinstance(header, fits.Header):
        header = fits.Header(header)

    hdu = fits.PrimaryHDU(data, header=header)

    # Create directories if required.
    if os.path.dirname(filename):
        os.makedirs(os.path.dirname(filename), mode=0o775, exist_ok=True)

    try:
        hdu.writeto(filename, **kwargs)
    except OSError as err:
        logger.error(f'Error writing image to {filename}: {err!r}')
    else:
        logger.debug(f'Image written to {filename}')
    finally:
        if exposure_event:
            exposure_event.set()


def update_observation_headers(file_path, info):
    """Update FITS headers with items from the Observation status.

    >>> # Check the headers
    >>> from panoptes.utils.images import fits as fits_utils
    >>> fits_fn = getfixture('unsolved_fits_file')
    >>> # Show original value
    >>> fits_utils.getval(fits_fn, 'FIELD')
    'KIC 8462852'

    >>> info = {'field_name': 'Tabbys Star'}
    >>> update_observation_headers(fits_fn, info)
    >>> # Show new value
    >>> fits_utils.getval(fits_fn, 'FIELD')
    'Tabbys Star'

    Args:
        file_path (str): Path to a FITS file.
        info (dict): The return dict from `pocs.observatory.Observation.status`,
            which includes basic information about the observation.
    """
    with fits.open(file_path, 'update') as f:
        hdu = f[0]
        hdu.header.set('IMAGEID', info.get('image_id', ''))
        hdu.header.set('SEQID', info.get('sequence_id', ''))
        hdu.header.set('FIELD', info.get('field_name', ''))
        hdu.header.set('RA-MNT', info.get('ra_mnt', ''), 'Degrees')
        hdu.header.set('HA-MNT', info.get('ha_mnt', ''), 'Degrees')
        hdu.header.set('DEC-MNT', info.get('dec_mnt', ''), 'Degrees')
        hdu.header.set('EQUINOX', info.get('equinox', 2000.))  # Assume J2000
        hdu.header.set('AIRMASS', info.get('airmass', ''), 'Sec(z)')
        hdu.header.set('FILTER', info.get('filter', ''))
        hdu.header.set('LAT-OBS', info.get('latitude', ''), 'Degrees')
        hdu.header.set('LONG-OBS', info.get('longitude', ''), 'Degrees')
        hdu.header.set('ELEV-OBS', info.get('elevation', ''), 'Meters')
        hdu.header.set('MOONSEP', info.get('moon_separation', ''), 'Degrees')
        hdu.header.set('MOONFRAC', info.get('moon_fraction', ''))
        hdu.header.set('CREATOR', info.get('creator', ''), 'POCS Software version')
        hdu.header.set('INSTRUME', info.get('camera_uid', ''), 'Camera ID')
        hdu.header.set('OBSERVER', info.get('observer', ''), 'PANOPTES Unit ID')
        hdu.header.set('ORIGIN', info.get('origin', ''))
        hdu.header.set('RA-RATE', info.get('tracking_rate_ra', ''), 'RA Tracking Rate')


def getdata(fn, *args, **kwargs):
    """Get the FITS data.

    Small wrapper around `astropy.io.fits.getdata` to auto-determine
    the FITS extension. This will return the data associated with the
    image.

    >>> fits_fn = getfixture('solved_fits_file')
    >>> d0 = getdata(fits_fn)
    >>> d0
    array([[2215, 2169, 2200, ..., 2169, 2235, 2168],
           [2123, 2191, 2133, ..., 2212, 2127, 2217],
           [2208, 2154, 2209, ..., 2159, 2233, 2187],
           ...,
           [2120, 2201, 2120, ..., 2209, 2126, 2195],
           [2219, 2151, 2199, ..., 2173, 2214, 2166],
           [2114, 2194, 2122, ..., 2202, 2125, 2204]], dtype=uint16)
    >>> d1, h1 = getdata(fits_fn, header=True)
    >>> (d0 == d1).all()
    True
    >>> h1['FIELD']
    'KIC 8462852'

    Args:
        fn (str): Path to FITS file.
        *args: Passed to `astropy.io.fits.getdata`.
        **kwargs: Passed to `astropy.io.fits.getdata`.

    Returns:
        `np.ndarray`: The FITS data.
    """
    return fits.getdata(fn, *args, **kwargs)


def getheader(fn, *args, **kwargs):
    """Get the FITS header.

    Small wrapper around `astropy.io.fits.getheader` to auto-determine
    the FITS extension. This will return the header associated with the
    image. If you need the compression header information use the astropy
    module directly.

    >>> fits_fn = getfixture('tiny_fits_file')
    >>> os.path.basename(fits_fn)
    'tiny.fits'
    >>> header = getheader(fits_fn)
    >>> header['IMAGEID']
    'PAN001_XXXXXX_20160909T081152'

    >>> # Works with fpacked files
    >>> fits_fn = getfixture('solved_fits_file')
    >>> os.path.basename(fits_fn)
    'solved.fits.fz'
    >>> header = getheader(fits_fn)
    >>> header['IMAGEID']
    'PAN001_XXXXXX_20160909T081152'

    Args:
        fn (str): Path to FITS file.
        *args: Passed to `astropy.io.fits.getheader`.
        **kwargs: Passed to `astropy.io.fits.getheader`.

    Returns:
        `astropy.io.fits.header.Header`: The FITS header for the data.
    """
    ext = 0
    if fn.endswith('.fz'):
        ext = 1
    return fits.getheader(fn, ext=ext)


def getwcs(fn, *args, **kwargs):
    """Get the WCS for the FITS file.

    Small wrapper around `astropy.wcs.WCS`.

    >>> from panoptes.utils.images import fits as fits_utils
    >>> fits_fn = getfixture('solved_fits_file')
    >>> wcs = fits_utils.getwcs(fits_fn)
    >>> wcs.is_celestial
    True
    >>> fits_fn = getfixture('unsolved_fits_file')
    >>> wcs = fits_utils.getwcs(fits_fn)
    >>> wcs.is_celestial
    False

    Args:
        fn (str): Path to FITS file.
        *args: Passed to `astropy.io.fits.getheader`.
        **kwargs: Passed to `astropy.io.fits.getheader`.

    Returns:
        `astropy.wcs.WCS`: The World Coordinate System information.
    """
    return WCS(getheader(fn, *args, **kwargs), *args, **kwargs)


def getval(fn, *args, **kwargs):
    """Get a value from the FITS header.

    Small wrapper around `astropy.io.fits.getval` to auto-determine
    the FITS extension. This will return the value from the header
    associated with the image (not the compression header). If you need
    the compression header information use the astropy module directly.

    >>> fits_fn = getfixture('tiny_fits_file')
    >>> getval(fits_fn, 'IMAGEID')
    'PAN001_XXXXXX_20160909T081152'

    Args:
        fn (str): Path to FITS file.

    Returns:
        str or float: Value from header (with no type conversion).
    """
    ext = 0
    if fn.endswith('.fz'):
        ext = 1
    return fits.getval(fn, *args, ext=ext, **kwargs)



import pytest

from panoptes.utils import error


def test_error(capsys, caplog):
    with pytest.raises(error.PanError) as e_info:
        raise error.PanError(msg='Testing message')

    assert str(e_info.value) == 'PanError: Testing message'

    with pytest.raises(error.PanError) as e_info:
        raise error.PanError()

    assert str(e_info.value) == 'PanError'

    with pytest.raises(SystemExit) as e_info:
        raise error.PanError(msg="Testing exit", exit=True)
    assert e_info.type == SystemExit
    assert caplog.records[-1].message == 'TERMINATING: Testing exit'

    with pytest.raises(SystemExit) as e_info:
        raise error.PanError(exit=True)
    assert e_info.type == SystemExit
    assert caplog.records[-1].message == 'TERMINATING: No reason specified'



import os
import pytest

import time
from datetime import timezone as tz
from datetime import datetime as dt
from astropy import units as u

from panoptes.utils import current_time
from panoptes.utils import CountdownTimer


def test_pretty_time():
    t0 = '2016-08-13 10:00:00'
    os.environ['POCSTIME'] = t0

    t1 = current_time(pretty=True)
    assert t1 == t0

    # This will increment one second - see docs
    t2 = current_time(flatten=True)
    assert t2 != t0
    assert t2 == '20160813T100001'

    # This will increment one second - see docs
    t3 = current_time(datetime=True)
    assert t3 == dt(2016, 8, 13, 10, 0, 2, tzinfo=tz.utc)


def test_countdown_timer_bad_input():
    with pytest.raises(ValueError):
        assert CountdownTimer('d')

    with pytest.raises(ValueError):
        assert CountdownTimer(current_time())

    with pytest.raises(AssertionError):
        assert CountdownTimer(-1)


def test_countdown_timer_non_blocking():
    timer = CountdownTimer(0)
    assert timer.is_non_blocking
    assert timer.time_left() == 0

    for arg, expected_duration in [(2, 2.0), (0.5, 0.5), (1 * u.second, 1.0)]:
        timer = CountdownTimer(arg)
        assert timer.duration == expected_duration


def test_countdown_timer():
    count_time = 1
    timer = CountdownTimer(count_time)
    assert timer.time_left() > 0
    assert timer.expired() is False
    assert timer.is_non_blocking is False

    counter = 0.
    while timer.time_left() > 0:
        time.sleep(0.1)
        counter += 0.1

    assert counter == pytest.approx(1)
    assert timer.time_left() == 0
    assert timer.expired() is True


def test_countdown_timer_sleep():
    count_time = 1
    timer = CountdownTimer(count_time)
    assert timer.time_left() > 0
    assert timer.expired() is False
    assert timer.is_non_blocking is False

    counter = 0.
    while timer.time_left() > 0.5:
        assert timer.sleep(max_sleep=0.1)
        counter += 0.1

    # Wait for the remaining half second
    assert timer.sleep() is False

    assert counter == pytest.approx(0.5)
    assert timer.time_left() == 0
    assert timer.expired() is True
    assert timer.sleep() is False



import tweepy
from ..logger import logger


class SocialTwitter(object):

    """Social Messaging sink to output to Twitter."""

    def __init__(self, **kwargs):
        consumer_key = kwargs.get('consumer_key', '')
        if consumer_key == '':
            raise ValueError('consumer_key parameter is not defined.')
        consumer_secret = kwargs.get('consumer_secret', '')
        if consumer_secret == '':
            raise ValueError('consumer_secret parameter is not defined.')
        access_token = kwargs.get('access_token', '')
        if access_token == '':
            raise ValueError('access_token parameter is not defined.')
        access_token_secret = kwargs.get('access_token_secret', '')
        if access_token_secret == '':
            raise ValueError('access_token_secret parameter is not defined.')

        # Output timestamp should always be True by default otherwise Twitter will reject duplicate statuses.
        self.output_timestamp = kwargs.get("output_timestamp", True)

        # Create a new twitter api object
        try:
            auth = tweepy.OAuthHandler(consumer_key, consumer_secret)
            auth.set_access_token(access_token, access_token_secret)

            self.api = tweepy.API(auth)
        except tweepy.TweepError:  # pragma: no cover
            msg = 'Error authenicating with Twitter. Please check your Twitter configuration.'
            logger.warning(msg)
            raise ValueError(msg)

    def send_message(self, msg, timestamp):
        try:
            # update_status returns a tweepy Status instance, but we
            # drop it on the floor because we don't have anything we
            # can do with it.
            if self.output_timestamp:
                self.api.update_status('{} - {}'.format(msg, timestamp))
            else:
                self.api.update_status(msg)
        except tweepy.TweepError:  # pragma: no cover
            logger.debug('Error tweeting message. Please check your Twitter configuration.')



import pytest

from astropy import units as u

from panoptes.utils import current_time
from panoptes.utils import serializers
from panoptes.utils import error


@pytest.fixture(scope='function')
def obj():
    return {
        "name": "Generic PANOPTES Unit",
        "pan_id": "PAN000",
        "location": {
            "name": "Mauna Loa Observatory",
            "latitude": 19.54 * u.degree,       # Astropy unit
            "longitude": "-155.58 deg",         # String unit
            "elevation": "3400.0 m",
            "horizon": 30 * u.degree,
            "flat_horizon": -6 * u.degree,
            "focus_horizon": -12 * u.degree,
            "observe_horizon": -18 * u.degree,
            "timezone": "US/Hawaii",
            "gmt_offset": -600,
        },
        "directories": {
            "base": "/var/panoptes",
            "images": "images",
            "data": "data",
            "resources": "POCS/resources/",
            "targets": "POCS/resources/targets",
            "mounts": "POCS/resources/mounts",
        },
        "db": {
            "name": "panoptes",
            "type": "file"
        },
        "empty": {},
        "current_time": current_time(),
        "bool": True,
        "exception": TypeError,
        "panoptes_exception": error.InvalidObservation
    }


def test_roundtrip_json(obj):
    config_str = serializers.to_json(obj)
    config = serializers.from_json(config_str)
    assert config['name'] == obj['name']
    assert config['location']['latitude'] == obj['location']['latitude']


def test_roundtrip_yaml(obj):
    config_str = serializers.to_yaml(obj)
    config = serializers.from_yaml(config_str)
    assert config['name'] == obj['name']
    assert config['location']['latitude'] == obj['location']['latitude']



import logging
from warnings import warn
from flask import Flask
from flask import request
from flask import jsonify
from flask.json import JSONEncoder

from multiprocessing import Process
from scalpl import Cut

from ..logger import logger
from . import load_config
from . import save_config
from ..serializers import _serialize_object

logging.getLogger('werkzeug').setLevel(logging.WARNING)

app = Flask(__name__)


class CustomJSONEncoder(JSONEncoder):

    def default(self, obj):
        return _serialize_object(obj)


app.json_encoder = CustomJSONEncoder


def config_server(host='localhost',
                  port=6563,
                  config_file=None,
                  ignore_local=False,
                  auto_save=False,
                  auto_start=True,
                  debug=False):
    """Start the config server in a separate process.

    A convenience function to start the config server.

    Args:
        host (str, optional): Name of host, default 'localhost'.
        port (int, optional): Port for server, default 6563.
        config_file (str|None, optional): The config file to load, defaults to
            `$PANDIR/conf_files/pocs.yaml`.
        ignore_local (bool, optional): If local config files should be ignored,
            default False.
        auto_save (bool, optional): If setting new values should auto-save to
            local file, default False.
        auto_start (bool, optional): If server process should be started
            automatically, default True.
        debug (bool, optional): Flask server debug mode, default False.

    Returns:
        `multiprocessing.Process`: The process running the config server.
    """
    app.config['auto_save'] = auto_save
    app.config['config_file'] = config_file
    app.config['ignore_local'] = ignore_local
    app.config['POCS'] = load_config(config_files=config_file, ignore_local=ignore_local)
    app.config['POCS_cut'] = Cut(app.config['POCS'])

    def start_server(**kwargs):
        try:
            app.run(**kwargs)
        except OSError:
            warn(f'Problem starting config server, is another config server already running?')
            return None

    server_process = Process(target=start_server,
                             kwargs=dict(host=host, port=port, debug=debug),
                             name='panoptes-config-server')

    if server_process is not None and auto_start:
        try:
            server_process.start()
        except KeyboardInterrupt:
            server_process.terminate()

    return server_process


@app.route('/get-config', methods=['GET', 'POST'])
def get_config_entry():
    """Get config entries from server.

    Endpoint that responds to GET and POST requests and returns
    configuration item corresponding to provided key or entire
    configuration. The key entries should be specified in dot-notation,
    with the names corresponding to the entries stored in the configuration
    file. See the [scalpl](https://pypi.org/project/scalpl/) documentation
    for details on the dot-notation.

    The endpoint should received a JSON document with a single key named "key"
    and a value that corresponds to the desired key within the configuration.

    For example, take the following configuration:

    ```
    { 
        'location': {
            'elevation': 3400.0,
        }
    }
    ```

    To get the corresponding value for the elevation, pass a JSON document similar to:

    ```
    '{"key": "location.elevation"}'
    ```
    """
    req_json = request.get_json()

    if request.is_json:
        # If requesting specific key
        try:
            key = req_json['key']
        except KeyError:
            return jsonify({
                'success': False,
                'msg': "No valid key found. Need json request: {'key': <config_entry>}"
            })

        if key is None:
            # Return all
            show_config = app.config['POCS']
        else:
            try:
                show_config = app.config['POCS_cut'].get(key, None)
            except KeyError:
                show_config = None
    else:
        # Return entire config
        show_config = app.config['POCS']

    return jsonify(show_config)


@app.route('/set-config', methods=['GET', 'POST'])
def set_config_entry():
    if request.is_json:
        req_data = request.get_json()

        try:
            app.config['POCS_cut'].update(req_data)
        except KeyError:
            for k, v in req_data.items():
                app.config['POCS_cut'].setdefault(k, v)

        # Config has been modified so save to file
        if app.config['auto_save'] and app.config['config_file'] is not None:
            save_config(app.config['config_file'], app.config['POCS_cut'].copy())

        return jsonify(req_data)

    return jsonify({
        'success': False,
        'msg': "Invalid. Need json request: {'key': <config_entry>, 'value': <new_values>}"
    })


@app.route('/reset-config', methods=['POST'])
def reset_config():
    if request.is_json:
        logger.warning(f'Resetting config server')
        req_data = request.get_json()

        if req_data['reset']:
            # Reload the config
            app.config['POCS'] = load_config(config_files=app.config['config_file'],
                                             ignore_local=app.config['ignore_local'])
            app.config['POCS_cut'] = Cut(app.config['POCS'])

        return jsonify(req_data)

    return jsonify({
        'success': False,
        'msg': "Invalid. Need json request: {'reset': True}"
    })



import os
import shutil

from astroplan import download_IERS_A
from astropy.utils import data

from ..logger import logger


class Downloader(object):
    """Downloads IERS Bulletin A and astrometry.net indices.

    IERS Bulletin A contains rapid determinations for earth orientation
    parameters, and is used by astroplan. Learn more at: https://www.iers.org

    Astrometry.net provides indices used to 'plate solve', i.e. to determine
    which stars are in an arbitrary image of the night sky.
    """

    def __init__(self,
                 data_folder=None,
                 wide_field=True,
                 narrow_field=False,
                 keep_going=True,
                 verbose=False):
        """
        Args:
            data_folder: Path to directory into which to copy the astrometry.net indices.
            wide_field: If True, downloads wide field astrometry.net indices.
            narrow_field: If True, downloads narrow field astrometry.net indices.
            keep_going: If False, exceptions are not suppressed. If True, returns False if there
                are any download failures, else returns True.
            verbose (bool, optional): If console output, default False.
        """
        self.data_folder = data_folder
        self.wide_field = wide_field
        self.narrow_field = narrow_field
        self.keep_going = keep_going
        self.verbose = verbose

    def download_all_files(self):
        """Downloads the files according to the attributes of this object."""
        result = True

        result = self.download_iers()

        if self.wide_field:
            for i in range(4110, 4119):
                if not self.download_one_file(f'4100/index-{i}.fits'):
                    result = False

        if self.narrow_field:
            for i in range(4210, 4219):
                if not self.download_one_file(f'4200/index-{i}.fits'):
                    result = False
        return result

    def download_iers(self):
        """Downloads the earth rotation catalog needed for accurate coordinate positions.

        Note: This download gives us a host of problems. This function should be called
        less than every 14 days otherwise a warning will be given. Currently we
        are downloading of our own google-based mirror of the data.

        Returns:
            TYPE: Description
        """
        try:
            download_IERS_A(show_progress=self.verbose)
            return True
        except Exception as e:
            logger.warning(f'Failed to download IERS A bulletin: {e}')
            return False

    def download_one_file(self, fn):
        """Downloads one astrometry.net file into self.data_folder."""
        dest = "{}/{}".format(self.data_folder, os.path.basename(fn))
        if os.path.exists(dest):
            return True
        url = "http://data.astrometry.net/{}".format(fn)
        try:
            df = data.download_file(url)
        except Exception as e:
            if not self.keep_going:
                raise e
            logger.warning(f'Failed to download {url}: {e}')
            return False
        # The file has been downloaded to some directory. Move the file into the data folder.
        try:
            self.create_data_folder()
            shutil.move(df, dest)
            return True
        except OSError as e:
            if not self.keep_going:
                raise e
            logger.warning(f"Problem saving {url}. Check permissions: {e}")
            return False

    def create_data_folder(self):
        """Creates the data folder if it does not exist."""
        if not os.path.exists(self.data_folder):
            logger.info("Creating data folder: {}.".format(self.data_folder))
            os.makedirs(self.data_folder)




import os
from contextlib import suppress
from warnings import warn

from ..logger import logger
from ..utils import listify
from ..serializers import from_yaml
from ..serializers import to_yaml


def load_config(config_files=None, simulator=None, parse=True, ignore_local=False):
    """Load configuation information

    This function supports loading of a number of different files. If no options
    are passed to `config_files` then the default `$PANDIR/conf_files/pocs.yaml`
    will be loaded. See Notes for additional information.

    .. note::

        The `config_files` parameter supports a number of options:
        * `config_files` is a list and loaded in order, so the first entry
        will have any values overwritten by similarly named keys in the second entry.
        * Entries can be placed in the `$PANDIR/conf_files` folder and
        should be passed as just the file name, e.g. [`weather.yaml`, `email.yaml`]
        for loading `$PANDIR/conf_files/weather.yaml` and `$PANDIR/conf_files/email.yaml`
        * The `.yaml` extension will be added if not present, so list can be
        written as just ['weather', 'email'].
        * `config_files` can also be specified by an absolute path, which can
        exist anywhere on the filesystem.
        * Local versions of files can override built-in versions and are automatically
        loaded if placed in the `$PANDIR/conf_files` folder. The files have a
        `<>_local.yaml` name, where `<>` is the built-in file. So a
        `$PANDIR/conf_files/pocs_local.yaml` will override any setting in the
        default `pocs.yaml` file.
        * Local files can be ignored (mostly for testing purposes) with the
        `ignore_local` parameter.

    Args:
        config_files (list, optional): A list of files to load as config,
            see Notes for details of how to specify files.
        simulator (list, optional): A list of hardware items that should be
            used as a simulator.
        parse (bool, optional): If the config file should attempt to create
            objects such as dates, astropy units, etc.
        ignore_local (bool, optional): If local files should be ignored, see
            Notes for details.

    Returns:
        dict: A dictionary of config items
    """

    # Default to the pocs.yaml file
    if config_files is None:
        config_files = ['pocs']
    config_files = listify(config_files)

    config = dict()

    config_dir = '{}/conf_files'.format(os.getenv('PANDIR'))

    for f in config_files:
        if not f.endswith('.yaml'):
            f = '{}.yaml'.format(f)

        if not f.startswith('/'):
            path = os.path.join(config_dir, f)
        else:
            path = f

        try:
            _add_to_conf(config, path, parse=parse)
        except Exception as e:
            warn(f"Problem with config file {path}, skipping. {e}")

        # Load local version of config
        if ignore_local is False:
            local_version = os.path.join(config_dir, f.replace('.', '_local.'))
            if os.path.exists(local_version):
                try:
                    _add_to_conf(config, local_version, parse=parse)
                except Exception:
                    warn(f"Problem with local config file {local_version}, skipping")

    # parse_config currently only corrects directory names.
    if parse:
        config = parse_config(config)

    return config


def save_config(path, config, overwrite=True):
    """Save config to local yaml file.

    This will save any entries into the `$PANDIR/conf_files/<path>_local.yaml` file to avoid
    clobbering what comes from the version control.

    Args:
        path (str): Path to save, can be relative or absolute. See Notes
            in `load_config`.
        config (dict): Config to save.
        overwrite (bool, optional): True if file should be updated, False
            to generate a warning for existing config. Defaults to True
            for updates.
    """
    # Make sure ends with '_local.yaml'
    base, ext = os.path.splitext(path)

    # Always want .yaml (although not actually used).
    ext = '.yaml'

    # Check for _local name.
    if not base.endswith('_local'):
        base = f'{base}_local'

    # Check full path location
    if not base.startswith('/'):
        config_dir = os.path.join(os.environ['PANDIR'], 'conf_files')
        base = os.path.join(config_dir, base)

    full_path = f'{base}{ext}'

    if os.path.exists(full_path) and overwrite is False:
        logger.warning(f"Path exists and overwrite=False: {full_path}")
    else:
        # Create directory if does not exist
        os.makedirs(os.path.dirname(full_path), exist_ok=True)
        with open(full_path, 'w') as f:
            print(config)
            to_yaml(config, stream=f)


def parse_config(config):
    """Parse the config dictionary for common objects.

    Currently only parses the following:
        * `directories` for relative path names.

    Args:
        config (dict): Config items.

    Returns:
        dict: Config items but with objects.
    """
    # Prepend the base directory to relative dirs
    if 'directories' in config:
        base_dir = os.getenv('PANDIR')
        for dir_name, rel_dir in config['directories'].items():
            abs_dir = os.path.normpath(os.path.join(base_dir, rel_dir))
            if abs_dir != rel_dir:
                config['directories'][dir_name] = abs_dir

    return config


def _add_to_conf(config, fn, parse=False):
    with suppress(IOError):
        with open(fn, 'r') as f:
            c = from_yaml(f, parse=parse)
            if c is not None and isinstance(c, dict):
                config.update(c)



"""The protocol_*.py files in this package are based on PySerial's file
test/handlers/protocol_test.py, modified for different behaviors.
The call serial.serial_for_url("XYZ://") looks for a class Serial in a
file named protocol_XYZ.py in this package (i.e. directory).
"""

from serial import serialutil


class NoOpSerial(serialutil.SerialBase):
    """No-op implementation of PySerial's SerialBase.

    Provides no-op implementation of various methods that SerialBase expects
    to have implemented by the sub-class. Can be used as is for a /dev/null
    type of behavior.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def in_waiting(self):
        """The number of input bytes available to read immediately."""
        return 0

    def open(self):
        """Open port.

        Raises:
            SerialException if the port cannot be opened.
        """
        self.is_open = True

    def close(self):
        """Close port immediately."""
        self.is_open = False

    def read(self, size=1):
        """Read size bytes.

        If a timeout is set it may return fewer characters than requested.
        With no timeout it will block until the requested number of bytes
        is read.

        Args:
            size: Number of bytes to read.

        Returns:
            Bytes read from the port, of type 'bytes'.

        Raises:
            SerialTimeoutException: In case a write timeout is configured for
                the port and the time is exceeded.
        """
        if not self.is_open:
            raise serialutil.portNotOpenError
        return bytes()

    def write(self, data):
        """
        Args:
            data: The data to write.

        Returns:
            Number of bytes written.

        Raises:
            SerialTimeoutException: In case a write timeout is configured for
                the port and the time is exceeded.
        """
        if not self.is_open:
            raise serialutil.portNotOpenError
        return 0

    def reset_input_buffer(self):
        """Remove any accumulated bytes from the device."""
        pass

    def reset_output_buffer(self):
        """Remove any accumulated bytes not yet sent to the device."""
        pass

    # --------------------------------------------------------------------------
    # There are a number of methods called by SerialBase that need to be
    # implemented by sub-classes, assuming their calls haven't been blocked
    # by replacing the calling methods/properties. These are no-op
    # implementations.

    def _reconfigure_port(self):
        """Reconfigure the open port after a property has been changed.

        If you need to know which property has been changed, override the
        setter for the appropriate properties.
        """
        pass

    def _update_rts_state(self):
        """Handle rts being set to some value.

        "self.rts = value" has been executed, for some value. This may not
        have changed the value.
        """
        pass

    def _update_dtr_state(self):
        """Handle dtr being set to some value.

        "self.dtr = value" has been executed, for some value. This may not
        have changed the value.
        """
        pass

    def _update_break_state(self):
        """Handle break_condition being set to some value.

        "self.break_condition = value" has been executed, for some value.
        This may not have changed the value.
        Note that break_condition is set and then cleared by send_break().
        """
        pass



import os
import pytest
import signal

from panoptes.utils import DelaySigTerm
from panoptes.utils import listify
from panoptes.utils import error
from panoptes.utils.library import load_module
from panoptes.utils.library import load_c_library


def test_bad_load_module():
    with pytest.raises(error.NotFound):
        load_module('FOOBAR')


def test_load_c_library():
    # Called without a `path` this will use find_library to locate libc.
    libc = load_c_library('c')
    assert libc._name[:4] == 'libc'

    libc = load_c_library('c', mode=None)
    assert libc._name[:4] == 'libc'


def test_load_c_library_fail():
    # Called without a `path` this will use find_library to locate libc.
    with pytest.raises(error.NotFound):
        load_c_library('foobar')


def test_listify():
    assert listify(12) == [12]
    assert listify([1, 2, 3]) == [1, 2, 3]


def test_empty_listify():
    assert listify(None) == []


def test_listfy_dicts():
    d = dict(a=42)

    d_vals = d.values()
    d_keys = d.keys()

    assert isinstance(listify(d_vals), list)
    assert listify(d_vals) == list(d_vals)

    assert isinstance(listify(d_keys), list)
    assert listify(d_keys) == list(d_keys)

    assert isinstance(listify(d), list)
    assert listify(d) == list(d_vals)


def test_delay_of_sigterm_with_nosignal():
    orig_sigterm_handler = signal.getsignal(signal.SIGTERM)

    with DelaySigTerm():
        assert signal.getsignal(signal.SIGTERM) != orig_sigterm_handler

    assert signal.getsignal(signal.SIGTERM) == orig_sigterm_handler


def test_delay_of_sigterm_with_handled_signal():
    """Confirm that another type of signal can be handled.

    In this test we'll send SIGCHLD, which should immediately call the
    signal_handler the test installs, demonstrating that only SIGTERM
    is affected by this DelaySigTerm.
    """
    test_signal = signal.SIGCHLD

    # Booleans to keep track of how far we've gotten.
    before_signal = False
    after_signal = False
    signal_handled = False
    after_with = False

    def signal_handler(signum, frame):
        assert before_signal

        nonlocal signal_handled
        assert not signal_handled
        signal_handled = True

        assert not after_signal

    old_test_signal_handler = signal.getsignal(test_signal)
    orig_sigterm_handler = signal.getsignal(signal.SIGTERM)
    try:
        # Install our handler.
        signal.signal(test_signal, signal_handler)

        with DelaySigTerm():
            assert signal.getsignal(signal.SIGTERM) != orig_sigterm_handler
            before_signal = True
            # Send the test signal. It should immediately
            # call our handler.
            os.kill(os.getpid(), test_signal)
            assert signal_handled
            after_signal = True

        after_with = True
        assert signal.getsignal(signal.SIGTERM) == orig_sigterm_handler
    finally:
        assert before_signal
        assert signal_handled
        assert after_signal
        assert after_with
        assert signal.getsignal(signal.SIGTERM) == orig_sigterm_handler
        signal.signal(test_signal, old_test_signal_handler)


def test_delay_of_sigterm_with_raised_exception():
    """Confirm that raising an exception inside the handler is OK."""
    test_signal = signal.SIGCHLD

    # Booleans to keep track of how far we've gotten.
    before_signal = False
    after_signal = False
    signal_handled = False
    exception_caught = False

    def signal_handler(signum, frame):
        assert before_signal

        nonlocal signal_handled
        assert not signal_handled
        signal_handled = True

        assert not after_signal
        raise UserWarning()

    old_test_signal_handler = signal.getsignal(test_signal)
    orig_sigterm_handler = signal.getsignal(signal.SIGTERM)
    try:
        # Install our handler.
        signal.signal(test_signal, signal_handler)

        with DelaySigTerm():
            assert signal.getsignal(signal.SIGTERM) != orig_sigterm_handler
            before_signal = True
            # Send the test signal. It should immediately
            # call our handler.
            os.kill(os.getpid(), test_signal)
            # Should not reach this point because signal_handler() should
            # be called because we called:
            #     signal.signal(other-handler, signal_handler)
            after_signal = True
            assert False, "Should not get here!"
    except UserWarning:
        assert before_signal
        assert signal_handled
        assert not after_signal
        assert not exception_caught
        assert signal.getsignal(signal.SIGTERM) == orig_sigterm_handler
        exception_caught = True
    finally:
        # Restore old handler before asserts.
        signal.signal(test_signal, old_test_signal_handler)

        assert before_signal
        assert signal_handled
        assert not after_signal
        assert exception_caught
        assert signal.getsignal(signal.SIGTERM) == orig_sigterm_handler


def test_delay_of_sigterm_with_sigterm():
    """Confirm that SIGTERM is in fact delayed."""

    # Booleans to keep track of how far we've gotten.
    before_signal = False
    after_signal = False
    signal_handled = False

    def signal_handler(signum, frame):
        assert before_signal
        assert after_signal

        nonlocal signal_handled
        assert not signal_handled
        signal_handled = True

    orig_sigterm_handler = signal.getsignal(signal.SIGTERM)
    try:
        # Install our handler.
        signal.signal(signal.SIGTERM, signal_handler)

        with DelaySigTerm():
            before_signal = True
            # Send SIGTERM. It should not call the handler yet.
            os.kill(os.getpid(), signal.SIGTERM)
            assert not signal_handled
            after_signal = True

        assert signal.getsignal(signal.SIGTERM) == signal_handler
        assert before_signal
        assert after_signal
        assert signal_handled
    finally:
        signal.signal(signal.SIGTERM, orig_sigterm_handler)



import numpy as np
from scipy import interpolate


class Horizon(object):
    """A simple class to define some coordinate points.

    Accepts a list of lists where each list consists of two points corresponding
    to an altitude (0-90) and an azimuth (0-360). If azimuth is a negative number
    (but greater than -360) then 360 will be added to put it in the correct
    range.

    The list are points that are obstruction points beyond the default horizon.
    """

    def __init__(self, obstructions=list(), default_horizon=30):
        """Create a list of horizon obstruction points.

        Each item in the `obstructions` list should be two or more points, where
        each point is an `[Alt, Az]` coordinate.

        Example:
            An example `obstruction_point` list::

                [
                [[40, 30], [40, 75]],   # From azimuth 30° to 75° there is an
                                        # obstruction that is at 40° altitude
                [[50, 180], [40, 200]], # From azimuth 180° to 200° there is
                                        # an obstruction that slopes from 50°
                                        # to 40° altitude
                ]

        Args:
            obstructions (list(list(list)), optional): A list of obstructions
                where each obstruction consists of a set of lists. The individual
                lists are alt/az pairs. Defaults to empty list in which case the
                `default_horizon` defines a flat horizon.
            default_horizon (float, optional): A default horizon to be used whenever
                there is no obstruction.

        """
        super().__init__()

        obstruction_list = list()
        for obstruction in obstructions:
            assert isinstance(obstruction, list), "Obstructions must be lists"
            assert len(obstruction) >= 2, "Obstructions must have at least 2 points"

            obstruction_line = list()
            for point in obstruction:
                assert isinstance(point, list), "Obstruction points must be lists"
                assert len(point) == 2, "Obstruction points must be 2 points"

                assert type(point[0]) is not bool, "Bool not allowed"
                assert type(point[1]) is not bool, "Bool not allowed"
                assert isinstance(point[0], (float, int, np.integer)), "Must be number-like"
                assert isinstance(point[1], (float, int, np.integer)), "Must be number-like"

                alt = float(point[0])
                az = float(point[1])

                assert 0. <= alt <= 90., "Altitude must be between 0-90 degrees"

                if az < 0.:
                    az += 360

                assert 0. <= az <= 360., "Azimuth must be between 0-360 degrees"
                obstruction_line.append((alt, az))
            obstruction_list.append(sorted(obstruction_line, key=lambda point: point[1]))

        self.obstructions = sorted(obstruction_list, key=lambda point: point[1])

        self.default_horizon = default_horizon
        self.horizon_line = np.ones(360) * self.default_horizon

        # Make helper lists of the alt and az
        self.alt = list()
        self.az = list()
        for obstruction in self.obstructions:
            self.alt.append([point[0] for point in obstruction])
            self.az.append([point[1] for point in obstruction])

        for obs_az, obs_alt in zip(self.az, self.alt):
            f = interpolate.interp1d(obs_az, obs_alt)

            x_range = np.arange(obs_az[0], obs_az[-1] + 1)
            new_y = f(x_range)

            # Assign over index elements
            for i, j in enumerate(x_range):
                self.horizon_line[int(j)] = new_y[i]



import pytest
import numpy as np
import random

from panoptes.utils.horizon import Horizon


def test_normal():
    hp = Horizon(obstructions=[
        [[20, 10], [40, 70]]
    ])
    assert isinstance(hp, Horizon)

    hp2 = Horizon(obstructions=[
        [[40, 45], [50, 50], [60, 45]]
    ])
    assert isinstance(hp2, Horizon)

    hp3 = Horizon()
    assert isinstance(hp3, Horizon)


def test_bad_length_tuple():
    with pytest.raises(AssertionError):
        Horizon(obstructions=[
            [[20], [40, 70]]
        ])


def test_bad_length_list():
    with pytest.raises(AssertionError):
        Horizon(obstructions=[
            [[40, 70]]
        ])


def test_bad_string():
    with pytest.raises(AssertionError):
        Horizon(obstructions=[
            [["x", 10], [40, 70]]
        ])


def test_too_many_points():
    with pytest.raises(AssertionError):
        Horizon(obstructions=[[[120, 60, 300]]])


def test_wrong_bool():
    with pytest.raises(AssertionError):
        Horizon(obstructions=[[[20, 200], [30, False]]])


def test_numpy_ints():
    range_length = 360
    points = [list(list(a) for a in zip(
        [random.randrange(15, 50) for _ in range(range_length)],  # Random height
        np.arange(1, range_length, 25)  # Set azimuth
    ))]
    points
    assert isinstance(Horizon(points), Horizon)


def test_negative_alt():
    with pytest.raises(AssertionError):
        Horizon(obstructions=[
            [[10, 20], [-1, 30]]
        ])


def test_good_negative_az():
    hp = Horizon(obstructions=[
        [[50, -10], [45, -20]]
    ])
    assert isinstance(hp, Horizon)

    hp2 = Horizon(obstructions=[
        [[10, -181], [20, -190]]
    ])
    assert isinstance(hp2, Horizon)


def test_bad_negative_az():
    with pytest.raises(AssertionError):
        Horizon(obstructions=[
            [[10, -361], [20, -350]]
        ])


def test_sorting():
    points = [
        [[10., 10.], [20., 20.]],
        [[30., 190.], [10., 180.]],
        [[10., 50.], [30., 60.]],
    ]
    hp = Horizon(obstructions=points)
    assert hp.obstructions == [[(10.0, 10.0), (20.0, 20.0)],
                               [(10.0, 50.0), (30.0, 60.0)],
                               [(10.0, 180.0), (30.0, 190.0)]]
